/*------------------------------------------------------------------------------
* crescent.c : hemisphere crescent receiver dependent functions
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2008/07/14 00:05:05 $
* history : 2008/05/21 1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"
#include "crescent.h"

#define NTOBS       4
#define CRESSYNC    "$BIN"      /* crescent bin sysc header */
#define CRESWAAS    80          /* crescent bin 80 msg : waas messages */
#define CRESIONUTC  94          /* crescent bin 94 msg : ion/utc parameters */
#define CRESEPH     95          /* crescent bin 95 msg : ephemeris */
#define CRESRAW     96          /* crescent bin 96 msg : raw phase and code */
#define U2(p)       (*((unsigned short *)(p)))
#define U4(p)       (*((unsigned int *)(p)))
#define I4(p)       (*((int *)(p)))
#define R8(p)       (*((double *)(p)))

static const char rcsid[]="$Id: crescent.c,v 1.2 2008/07/14 00:05:05 TTAKA Exp $";

/* output rinex obs types */
static char tobs[NTOBS][3]={"C1","L1","D1","S1"};

/* application defined functions ---------------------------------------------*/
extern int showmsg(char *format,...);

/* check checksum ------------------------------------------------------------*/
static int chksum(const unsigned char *buff, int len)
{
    int i,stat;
    unsigned short sum=0;
    
    for (i=8;i<len-4;i++) sum+=buff[i];
    if (!(stat=((sum>>8)==buff[len-3])&&((sum&0xFF)==buff[len-4])&&
                buff[len-2]==0x0D&&buff[len-1]==0x0A))
        fprintf(stderr,"chek sum : sum=%02X%02X %02X%02X %02X%02X\n",
                sum>>8,sum&0xFF,buff[len-3],buff[len-4],buff[len-2],buff[len-1]);
    return stat;
}
/* decode bin96 raw phase and code -------------------------------------------*/
static int decoderawobs(raw_t *raw, const unsigned char *buff, int len, int opt,
                        obsd_t *obs)
{
    gtime_t time;
    double tow,t,off;
    int i,n,sat,week,word2,val;
    unsigned int word1,sc;
    
    week=U2(buff+2);
    tow=R8(buff+4);
    t=floor(tow*1000.0+0.5)/1000.0; /* round by 1ms */
    off=CLIGHT*(t-tow); /* time tag offset (m) */
    time=gpst2time(week,t);
    raw->time=time;
    val=screent(time,raw->ts,raw->te,raw->tint);
    
    for (i=n=0;i<12&&n<MAXOBS;i++) {
        word1=U4(buff+12+i*24);
        if (!(sat=word1&0xFF)||sat>MAXSAT) continue;
        word2=I4(buff+16+i*24);
        obs[n].time=time;
        obs[n].sat=sat;
        obs[n].L[0]=word2&1?(R8(buff+28+i*24)-off)*FREQ1/CLIGHT:0.0;
        if (!(word2&1)) obs[n].L[0]=0.0;
        obs[n].P[0]=R8(buff+20+i*24)-off;
        obs[n].D[0]=(float)(word2/16/4096.0);
        obs[n].SNR[0]=(unsigned char)(10.0*log10(0.8192*((word1>>8)&0xFF))+30.0);
        sc=(unsigned int)(word1>>24);
        obs[n].LLI[0]=(int)((unsigned char)sc-(unsigned char)raw->lockc[SAT(sat)-1][0])>0;
        if (val) raw->lockc[SAT(sat)-1][0]=sc;
        n++;
    }
    return val?n:0;
}
/* decode bin 95 ephemeris ---------------------------------------------------*/
static int decoderaweph(const unsigned char *buff, int len, eph_t *ephs)
{
    eph_t eph={0};
    double ttr,tow,ion[8],utc[4];
    unsigned int words[10];
    int i,j,sat,week,leaps,iode1=-1,iode2=-1;
    
    sat=U2(buff);
    if (len!=128||sat<=0||32<sat) return 0;
    tow=U4(buff+4)*6.0;
    eph.sat=(unsigned char)sat;
    week=(U4(buff+16)>>20)+OFFWEEK;
    eph.ttr=gpst2time(week,tow);
    for (i=0;i<3;i++) {
        for (j=0;j<10;j++) words[j]=U4(buff+8+i*40+j*4)>>6;
        if (decodefrm(words,&eph,ion,utc,&leaps)!=i+1) return 0;
        if      (i==0) iode1=eph.iodc&0xFF;
        else if (i==1) iode2=eph.iode;
        else if (eph.iode!=iode1||eph.iode!=iode2) return 0;
    }
    ttr=time2gpst(eph.ttr,&week);
    if (ttr>302400.0&&eph.toes<302400.0) {
        eph.ttr=gpst2time(++week,ttr-604800.0);
    }
    eph.toe=gpst2time(week,eph.toes);   
    ephs[sat-1]=eph;
    return sat;
}
/* decode bin 94 ion/utc parameters ------------------------------------------*/
static int decodeionutc(const unsigned char *buff, int len, double *ion,
                        double *utc, int *leaps)
{
    int i;
    
    if (len!=96) return 0;
    for (i=0;i<8;i++) ion[i]=R8(buff+i*8);
    utc[0]=R8(buff+64);
    utc[1]=R8(buff+72);
    utc[2]=(double)U4(buff+80);
    utc[3]=(double)U2(buff+84);
    *leaps=I4(buff+90);
    return 1;
}
/* decode bin 80 waas messages -----------------------------------------------*/
static int decoderawwaas(gtime_t time, const unsigned char *buff, int len,
                         sbsmsg_t *sbsmsg)
{
    double tow;
    unsigned int word;
    int i,j,k,sat;
    
    sat=U2(buff);
    if (len!=40||sat<MINPRNSBS||MAXPRNSBS<sat||time.time==0) return 0;
    sbsmsg->sat=sat;
    sbsmsg->tow=U4(buff+4);
    tow=time2gpst(time,&sbsmsg->week);
    if (fabs(tow-sbsmsg->tow)>302400.0) {
        if      (tow<302400.0) sbsmsg->week--;
        else if (tow>302400.0) sbsmsg->week++;
    }
    for (i=k=0;i<8&&k<29;i++) {
        word=U4(buff+8+i*4);
        for (j=0;j<4&&k<29;j++) sbsmsg->msg[k++]=(unsigned char)(word>>(3-j)*8);
    }
    sbsmsg->msg[28]&=0xC0;
    return sat;
}
/* decode crescent bin message -------------------------------------------------
* decode and input a crescent bin message. support bin 80, 94, 95, 96.
* args   : raw_t *raw    IO     raw decode control struct
*          unsigned char *buff I message buffer
*          int   len     I      buffer length (bytes)
*          int   opt     I      decode option
*          obs_t *obs    IO     observation data (NULL: no output)
*          nav_t *nav    IO     navigation message (NULL: no output)
*          sbsmsg_t *sbsmsg IO  sbas message (NULL: no output)
*          int   *sat    O      satellite prn number for ephemeris or sbas msg
* return : input status (-1:error, 0:no valid data, 1:obs data, 2:ephemeris,
*                        3:sbas message, 9:ion/utc parameters)
*-----------------------------------------------------------------------------*/
extern int decodecres(raw_t *raw, const unsigned char *buff, int len, int opt,
                      obs_t *obs, nav_t *nav, sbsmsg_t *sbsmsg, int *sat)
{
    if (!chksum(buff,len)) return -1;
    
    switch (U2(buff+4)) {
    case CRESRAW:
        if (obs&&(obs->n=decoderawobs(raw,buff+8,len-12,opt,obs->data))>0) {
            return 1;
        }
        break;
    case CRESEPH:
        if (nav&&(*sat=decoderaweph(buff+8,len-12,nav->eph))>0) {
            return screent(nav->eph[*sat-1].ttr,raw->ts,raw->te,0.0)?2:0;
        }
        break;
    case CRESWAAS:
        if (sbsmsg&&(*sat=decoderawwaas(raw->time,buff+8,len-12,sbsmsg))>0) {
            return screent(gpst2time(sbsmsg->week,sbsmsg->tow),raw->ts,raw->te,0.0)?3:0;
        }
        break;
    case CRESIONUTC:
        if (nav&&(decodeionutc(buff+8,len-12,nav->ion,nav->utc,&nav->leaps))) return 9;
        break;
    }
    return 0;
}
/* convert cresent log file ----------------------------------------------------
* convert cresent log file to rinex obs/nav and sbas message file. log file
* shall include crescent bin messages bin 80, 94, 95, 96.
* args   : FILE  *fp     I      input log file pointer
*          FILE  *ofp    I      output rinex obs file pointer (NULL:no output)
*          FILE  *nfp    I      output rinex nav file pointer (NULL:no output)
*          FILE  *sfp    I      output sbas message file pointer (NULL:no output)
*          gtime_t ts    I      time start
*          gtime_t te    I      time end
*          double tint   I      time interval
*          rnxopt_t *rnxopt I   rinex options
*          int    opt    I      options
* return : none
*-----------------------------------------------------------------------------*/
extern void convcres(FILE *fp, FILE *ofp, FILE *nfp, FILE *sfp, gtime_t ts,
                     gtime_t te, double tint, const rnxopt_t *rnxopt, int opt)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obs={0};
    nav_t nav={0};
    sbsmsg_t sbsmsg;
    raw_t raw={{0}};
    gtime_t time={0},tstart={0};
    int i,c,sat,type,nmsg[4]={0};
    unsigned short len;
    unsigned char buff[1024];
    char s[128]="",s1[64],s2[64];
    char *sync=CRESSYNC;
    
    obs.data=data; nav.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
        
    while ((c=fgetc(fp))!=EOF) {
        buff[0]=(unsigned char)c;
        if (buff[0]!=sync[0]) continue;
        if (fread(buff+1,3,1,fp)<1) break;
        for (i=1;i<4;i++) if (buff[i]!=sync[i]) break;
        if (i<4) continue;
        if (fread(buff+4,4,1,fp)<1) break;
        if ((len=U2(buff+6))>sizeof(buff)-12) {nmsg[3]++; continue;}
        if (fread(buff+8,len+4,1,fp)<1) break;
        
        type=decodecres(&raw,buff,len+12,opt,&obs,&nav,&sbsmsg,&sat);
        
        if (type==1&&ofp) {
            if (nmsg[0]==0) {
                outrnxobsh(ofp,rnxopt,tobs,NTOBS,obs.data[0].time);
            }
            outrnxobsb(ofp,obs.data,obs.n,0,tobs,NTOBS);
            nmsg[0]++;
            time=obs.data[0].time;
        }
        else if (type==2&&nfp) {
            if (nmsg[1]==0) {
                outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
            }
            outrnxnavb(nfp,&nav.eph[sat-1]);
            nmsg[1]++;
            if (!ofp) time=nav.eph[sat-1].ttr;
        }
        else if (type==3&&sfp) {
            sbsoutmsg(sfp,&sbsmsg);
            nmsg[2]++;
            if (!ofp&&!nfp) time=gpst2time(sbsmsg.week,sbsmsg.tow);
        }
        else if (type==-1) nmsg[3]++;
        if (te.time!=0&&timediff(obs.data[0].time,te)>DTTOL+1.0) break;
        if (time.time!=0) {
            time2str(time,s1,0); sprintf(s,"%s: ",s1);
            if (tstart.time==0) tstart=time;
        }
        if (showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],
                    nmsg[2],nmsg[3])) break;
    }
    if (nfp) {
        rewind(nfp);
        outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
    }
    if (tstart.time!=0) {
        time2str(tstart,s1,0); time2str(time,s2,0);
        sprintf(s,"%s-%s: ",s1,s2+5);
    }
    showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],nmsg[2],nmsg[3]);
}
/* read crescent log file ------------------------------------------------------
* convert cresent log file to rinex obs/nav and sbas message file. log file
* shall include crescent bin messages bin 80, 94, 95, 96.
* args   : FILE  *fp     I      input file pointer
*          gtime_t ts,te I      time start/end
*          double tint   I      time interval
*          int   opt     I      read option)
*          obs_t *obs    IO     observation data    (NULL: no output)
*          nav_t *nav    IO     navigation messages (NULL: no output)
*          sbs_t *sbs    IO     sbas messages       (NULL: no output)
*          int   *nmax   I      max number of messages 
* return : none
*-----------------------------------------------------------------------------*/
extern void readcres(FILE *fp, gtime_t ts, gtime_t te, double tint, int opt,
                     obs_t *obs, nav_t *nav, sbs_t *sbs, int *nmax)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obsp={0};
    nav_t navp={0};
    sbsmsg_t sbsp;
    raw_t raw={{0}};
    int i,c,sat,type;
    unsigned short len;
    unsigned char buff[1024];
    char *sync=CRESSYNC;
    
    obsp.data=data; navp.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        buff[0]=(unsigned char)c;
        if (buff[0]!=sync[0]) continue;
        if (fread(buff+1,3,1,fp)<1) break;
        for (i=1;i<4;i++) if (buff[i]!=sync[i]) break;
        if (i<4) continue;
        if (fread(buff+4,4,1,fp)<1) break;
        if ((len=U2(buff+6))>sizeof(buff)-12) continue;
        if (fread(buff+8,len+4,1,fp)<1) break;
        
        type=decodecres(&raw,buff,len+12,opt,&obsp,&navp,&sbsp,&sat);
        
        if (type==1&&obs) {
            addobs(obs,&obsp,nmax);
        }
        else if (type==2&&nav) {
            addnav(nav,&navp,sat,nmax+1);
        }
        else if (type==3&&sbs) {
            addsbs(sbs,sbsp,nmax+2);
        }
        else if (type==9&&nav) {
            addionutc(nav,&navp);
        }
        if (te.time!=0&&timediff(obsp.data[0].time,te)>DTTOL) break;
    }
}
