/*------------------------------------------------------------------------------
* ublox.c : ublox receiver dependent functions
*
*          Copyright (C) 2007,2008 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2008/07/14 00:05:05 $
* history : 2007/10/08 1.0 new
*           2008/06/16 1.1 separate common functions to rcvcmn.c
*-----------------------------------------------------------------------------*/
#include "rtklib.h"
#include "ublox.h"

#define NTOBS       4
#define UBXSYN      0xB562      /* ubx message start sync code */
#define RXMRAW      0x0210      /* ubx message id : raw observation */
#define RXMSFRB     0x0211      /* ubx message id : nav data subframe */
#define U1(p)       (*((unsigned char *)(p)))
#define I1(p)       (*((char *)(p)))
#define U2(p)       (*((unsigned short *)(p)))
#define U4(p)       (*((unsigned int *)(p)))
#define R8(p)       (*((double *)(p)))
#define R4(p)       (*((float *)(p)))

static const char rcsid[]="$Id: ublox.c,v 1.2 2008/07/14 00:05:05 TTAKA Exp $";

/* output rinex obs types */
static char tobs[NTOBS][3]={"L1","C1","D1","S1"};

/* application defined functions ---------------------------------------------*/
extern int showmsg(char *format,...);

/* decode ubx raw observation data -------------------------------------------*/
static int decodeubxraw(raw_t *raw, const unsigned char *buff, int len,
                        obsd_t *obs, int opt)
{
    gtime_t time;
    double tow;
    int i,n=0,nsv,week,val;
    
    week=U2(buff+4);
    tow=U4(buff)*0.001;
    nsv=U1(buff+6);
    if (week>=4096||tow>604800.0||len<8+24*nsv) return 0;
    time=gpst2time(week,tow);
    val=screent(time,raw->ts,raw->te,raw->tint);
    raw->time=time;
    for (i=0;i<nsv&&i<MAXOBS;i++) {
        obs[n].time=time;
        obs[n].L[0]=R8(buff+ 8+24*i);
        obs[n].L[1]=0.0;
        if (opt&OPTUBX5) obs[n].L[0]*=-1.0; /* flip sign if ublox-5 */
        obs[n].P[0]=R8(buff+16+24*i);
        obs[n].P[1]=0.0;
        obs[n].D[0]=R4(buff+24+24*i);
        obs[n].D[1]=0.0;
        obs[n].sat=U1(buff+28+24*i);
        obs[n].SNR[0]=I1(buff+30+24*i);
        obs[n].SNR[1]=0;
        obs[n].LLI[0]=U1(buff+31+24*i);
        obs[n].LLI[1]=0;
        if (obs[n].LLI[0]&0x1) raw->lockc[SAT(obs[n].sat)-1][0]=1;
        if (val) {
            if (raw->lockc[SAT(obs[n].sat)-1][0]) obs[n].LLI[0]|=1;
            raw->lockc[SAT(obs[n].sat)-1][0]=0;
        }
        n++;
    }
    return val?n:0;
}
/* decode ublox messages -------------------------------------------------------
* decode and input a ublox message. support RXMRAW and RXMSFRB.
* args   : raw_t *raw    IO     raw decode control struct
*          unsigned char *buff I message buffer
*          int   len     I      buffer length (bytes)
*          int   opt     I      decode option
*          obs_t *obs    IO     observation data (NULL: no output)
*          nav_t *nav    IO     navigation message (NULL: no output)
*          sbsmsg_t *sbsmsg IO  sbas message (NULL: no output)
*          int   *sat    O      satellite prn number for ephemeris or sbas msg
* return : input status (-1:error, 0:no valid data, 1:obs data, 2:ephemeris,
*                        3:sbas message, 9:ion/utc parameters)
*-----------------------------------------------------------------------------*/
extern int decodeubx(raw_t *raw, const unsigned char *buff, int len, int opt,
                     obs_t *obs, nav_t *nav, sbsmsg_t *sbsmsg, int *sat)
{
    unsigned int words[10];
    int i,type;
    unsigned char cka=0,ckb=0;
    
    for (i=0;i<len-2;i++) { /* check sum */
        cka+=buff[i]; ckb+=cka;
    }
    if (cka!=buff[len-2]||ckb!=buff[len-1]) return -1;
    
    switch (((unsigned short)buff[0]<<8)+buff[1]) {
    case RXMRAW:
        if (obs&&(obs->n=decodeubxraw(raw,buff+4,len-6,obs->data,opt))>0) {
            return 1;
        }
        break;
    case RXMSFRB:
        if (len-6!=42) break;
        *sat=U1(buff+5);
        for (i=0;i<10;i++) words[i]=U4(buff+6+4*i);
        if (1<=*sat&&*sat<=MAXPRNGPS) { /* gps */
            if (!nav) return 0;
            type=decodenav(raw,*sat,words,nav);
            if (type!=2) return type;
            return screent(nav->eph[*sat-1].ttr,raw->ts,raw->te,0.0)?2:0;
        }
        else if (MINPRNSBS<=*sat&&*sat<=MAXPRNSBS) { /* sbas */
            if (sbsmsg&&sbsdecodemsg(obs->data[0].time,*sat,words,sbsmsg)) {
                return screent(gpst2time(sbsmsg->week,sbsmsg->tow),raw->ts,raw->te,0.0)?3:0;
            }
        }
        break;
    }
    return 0;
}
/* convert ublox log file ------------------------------------------------------
* convert ublox log file to rinex obs/nav and sbas message file. log file shall
* contain RXMRAW and RXMSFRB.
* args   : FILE  *fp     I      input log file pointer
*          FILE  *ofp    I      output rinex obs file pointer (NULL:no output)
*          FILE  *nfp    I      output rinex nav file pointer (NULL:no output)
*          FILE  *sfp    I      output sbas message file pointer (NULL:no output)
*          gtime_t ts    I      time start
*          gtime_t te    I      time end
*          double tint   I      time interval
*          int   opt     I      conversion option
*          const char *file I   input file
* return : none
*-----------------------------------------------------------------------------*/
extern void convubx(FILE *fp, FILE *ofp, FILE *nfp, FILE *sfp, gtime_t ts,
                    gtime_t te, double tint, const rnxopt_t *rnxopt, int opt)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obs={0};
    nav_t nav={0};
    sbsmsg_t sbsmsg;
    raw_t raw={{0}};
    gtime_t time={0},tstart={0};
    int c,sat,type,nmsg[4]={0};
    unsigned short sync=0,len;
    unsigned char buff[2048];
    char s[128]="",s1[64],s2[64];
    
    obs.data=data; nav.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if ((sync=(sync<<8)+(unsigned char)c)!=UBXSYN) continue;
        if (fread(buff,4,1,fp)<1) break;
        len=buff[2]+((unsigned short)buff[3]<<8);
        if ((int)len>sizeof(buff)-6) {nmsg[3]++; continue;}
        if ((int)fread(buff+4,1,len+2,fp)<len+2) break;
        
        type=decodeubx(&raw,buff,len+6,opt,&obs,&nav,&sbsmsg,&sat);
        
        if (type==1&&ofp) {
            if (nmsg[0]==0) {
                outrnxobsh(ofp,rnxopt,tobs,NTOBS,obs.data[0].time);
            }
            outrnxobsb(ofp,obs.data,obs.n,0,tobs,NTOBS);
            nmsg[0]++;
            time=obs.data[0].time;
        }
        else if (type==2&&nfp) {
            if (nmsg[1]==0) {
                outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
            }
            outrnxnavb(nfp,&nav.eph[sat-1]);
            nmsg[1]++;
            if (!ofp) time=nav.eph[sat-1].ttr;
        }
        else if (type==3&&sfp) {
            sbsoutmsg(sfp,&sbsmsg);
            nmsg[2]++;
            if (!ofp&&!nfp) time=gpst2time(sbsmsg.week,sbsmsg.tow);
        }
        else if (type==-1) nmsg[3]++;
        if (te.time!=0&&timediff(obs.data[0].time,te)>DTTOL+1.0) break;
        if (time.time!=0) {
            time2str(time,s1,0); sprintf(s,"%s: ",s1);
            if (tstart.time==0) tstart=time;
        }
        if (showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],
                    nmsg[2],nmsg[3])) break;
    }
    if (nfp) {
        rewind(nfp);
        outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
    }
    if (tstart.time!=0) {
        time2str(tstart,s1,0); time2str(time,s2,0);
        sprintf(s,"%s-%s: ",s1,s2+5);
    }
    showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],nmsg[2],nmsg[3]);
}
/* read ublox log file ----------------------------------------------------------
* read ublox log file as obs data, nav message and sbas messages. log file shall
* contain RXMRAW and RXMSFRB.
* args   : FILE  *fp     I      input file pointer
*          gtime_t ts,te I      time start/end
*          double tint   I      time interval
*          int   opt     I      read option)
*          obs_t *obs    IO     observation data    (NULL: no output)
*          nav_t *nav    IO     navigation messages (NULL: no output)
*          sbs_t *sbs    IO     sbas messages       (NULL: no output)
*          int   *nmax   I      max number of messages 
* return : none
*-----------------------------------------------------------------------------*/
extern void readubx(FILE *fp, gtime_t ts, gtime_t te, double tint, int opt,
                    obs_t *obs, nav_t *nav, sbs_t *sbs, int *nmax)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obsp={0};
    nav_t navp={0};
    sbsmsg_t sbsp;
    raw_t raw={{0}};
    int c,sat,type;
    unsigned short sync=0,len;
    unsigned char buff[2048];
    
    obsp.data=data; navp.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if ((sync=(sync<<8)+(unsigned char)c)!=UBXSYN) continue;
        if (fread(buff,4,1,fp)<1) break;
        len=buff[2]+((unsigned short)buff[3]<<8);
        if ((int)len>sizeof(buff)-6) continue;
        if ((int)fread(buff+4,1,len+2,fp)<len+2) break;
        
        type=decodeubx(&raw,buff,len+6,opt,&obsp,&navp,&sbsp,&sat);
        
        if (type==1&&obs) {
            addobs(obs,&obsp,nmax);
        }
        else if (type==2&&nav) {
            addnav(nav,&navp,sat,nmax+1);
        }
        else if (type==3&&sbs) {
            addsbs(sbs,sbsp,nmax+2);
        }
        else if (type==9&&nav) {
            addionutc(nav,&navp);
        }
        if (te.time!=0&&timediff(obsp.data[0].time,te)>DTTOL) break;
    }
}
