/*------------------------------------------------------------------------------
* rinex.c : input and output rinex observation data/navigation messages
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2006/01/16 1.0 new
*           2007/03/14 1.1 read P1 if no obstype of C1
*           2007/04/27 1.2 add readrnxt() function
*           2007/05/25 1.3 add support of file path with wild-card (*)
*                          add support of compressed files
*           2007/11/02 1.4 support sbas/geo satellite
*                          support doppler observables
*                          support rinex bug of week handover
*                          add rinex obs/nav output functions
*           2008/06/16 1.5 export readrnxf(), add compress()
*                          separate sortobs(), uniqeph(), screent()
*           2008/10/28 1.6 fix bug on reading rinex obs header types of observ
*-----------------------------------------------------------------------------*/
#include <ctype.h>
#include "rtklib.h"

static const char rcsid[]="$Id: rinex.c,v 1.1 2008/07/17 21:48:06 ttaka Exp ttaka $";

/* constants -----------------------------------------------------------------*/

#define RNXVER      2.1             /* rinex format version */
#define MAXFILES    100             /* max number of expanded files */

const char *obstypes[NFREQ*4]={"L1","L2","C1","P2","D1","D2","S1","S2"};
                                    /* read obs types */

/* read rinex header ---------------------------------------------------------*/
static char readrnxh(FILE *fp, char tobs[][3], int *nt, double *ion,
                     double *utc, int *leaps)
{
    int i,j,n;
    char buff[256],trnx=' ';
    
    trace(3,"readrnxh:\n");
    
    *nt=0;
    while (fgets(buff,sizeof(buff),fp)) {
        if (strlen(buff)<=60) continue;
        else if (strstr(buff+60,"RINEX VERSION / TYPE")) {
            trnx=*(buff+20);
        }
        else if (strstr(buff+60,"# / TYPES OF OBSERV")) {
            n=(int)str2num(buff,0,6);
            for (i=0,j=10;i<n;i++,j+=6,(*nt)++) {
                if (j>58) {if (!fgets(buff,sizeof(buff),fp)) break; j=10;}
                strncpy(tobs[i],buff+j,2); tobs[i][2]='\0';
            }
        }
        else if (strstr(buff+60,"ION ALPHA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"ION BETA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i+4]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"DELTA-UTC: A0,A1,T,W")) {
            for (i=0,j=3;i<2;i++,j+=19) utc[i]=str2num(buff,j,19);
            for (;i<4;i++,j+=9) utc[i]=str2num(buff,j,9);
        }
        else if (strstr(buff+60,"LEAP SECONDS")) {
            *leaps=(int)str2num(buff,0,6);
        }
        else if (strstr(buff+60,"END OF HEADER")) break;
    }
    return trnx;
}
/* read rinex observation data -----------------------------------------------*/
static int readrnxo(FILE *fp, gtime_t ts, gtime_t te, double tint,
                    char tobs[][3], int nt, int rcv, obs_t *obs, int *osiz)
{
    gtime_t time;
    obsd_t *data;
    double value;
    int i,j,k,n=0,m,event=0,index[32]={0},sats[128];
    char buff[256];
    unsigned char lock;
    
    trace(3,"readrnxo: nt=%d osiz=%d\n",nt,*osiz);
    
    for (i=0;i<NFREQ*4;i++) {
        for (j=0;j<nt;j++) if (!strcmp(tobs[j],obstypes[i])) {index[j]=i+1; break;}
        if (j>=nt&&!strcmp(obstypes[i],"C1")) { /* read P1 instead of C1 */
            for (j=0;j<nt;j++) if (!strcmp(tobs[j],"P1")) index[j]=i+1;
        }
    }
    i=-1;
    while (fgets(buff,sizeof(buff),fp)) {
        if (i<0) {
            n=(int)str2num(buff,29,3);
            if ((event=(int)str2num(buff,27,2))==4) ; /* header info follows */
            else if (!str2time(buff,0,26,&time)) {
                for (j=0,k=32;j<n;j++,k+=3) {
                    if (k>=68) {if (!fgets(buff,sizeof(buff),fp)) break; k=32;}
                    sats[j]=(unsigned char)str2num(buff,k+1,2);
                    switch (*(buff+k)) {
                    case ' ': case 'G':     break; /* gps */
                    case 'S': sats[j]+=100; break; /* sbas/geo */
                    case 'R': sats[j]+=50;  break; /* glonass */
                    default : sats[j]=0;    break; /* other */
                    }
                }
            }
            else continue;
        }
        else if (event!=4&&sats[i]>=0) {
            if (*osiz<=obs->n) {
                if (*osiz==0) *osiz=65532; else *osiz+=65532;
                if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*(*osiz)))) {
                    trace(1,"readrnxo: memory allocation error (n=%d)\n",*osiz);
                    *osiz=obs->n=0; return -1;
                }
            }
            data=&obs->data[obs->n];
            data->time=time; data->sat=sats[i]; data->rcv=(unsigned char)rcv;
            for (j=0;j<NFREQ;j++) {
                data->L[j]=data->P[j]=data->D[j]=0.0;
                data->SNR[j]=data->LLI[j]=0;
            }
            for (j=k=0;j<nt;j++,k+=16) {
                if (k>=80) {if (!fgets(buff,sizeof(buff),fp)) break; k=0;}
                value=str2num(buff,k,14); if (fabs(value)<1E-6) continue;
                lock=(unsigned char)str2num(buff,k+14,1)&3; /* 1:loss of lock */
                                                            /* 2:halfcycle amb */
                if (!index[j]) continue;
                m=(index[j]-1)%NFREQ;
                switch ((index[j]-1)/NFREQ) {
                case 0: data->L[m]=value; data->LLI[m]=lock; break;
                case 1: data->P[m]=value; break;
                case 2: data->D[m]=(float)value; break;
                case 3: data->SNR[m]=(unsigned char)value; break;
                }
            }
            if (screent(time,ts,te,tint)) obs->n++;
        }
        if (++i>=n) i=-1;
    }
    trace(3,"readrnxo: nobs=%d\n",obs->n);
    return 0;
}
/* adjust time considering week handover -------------------------------------*/
static gtime_t adjweek(gtime_t t, gtime_t t0)
{
    double tt=timediff(t,t0);
    if (tt<-302400.0) return timeadd(t, 604800.0);
    if (tt> 302400.0) return timeadd(t,-604800.0);
    return t;
}
/* read rinex navigation message ---------------------------------------------*/
static int readrnxn(FILE *fp, nav_t *nav, int *nsiz)
{
    gtime_t toc;
    eph_t *eph;
    double e[64];
    int i=0,j,sat=1;
    char buff[256];
    
    trace(3,"readrnxn: nsiz=%d\n",*nsiz);
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (i==0) {
            sat=(int)str2num(buff,0,2);
            if (str2time(buff,3,17,&toc)) continue;
        }
        for (j=i==0?22:3;j<=60;j+=19) e[i++]=str2num(buff,j,19);
        if (i>=31) {
            if (*nsiz<=nav->n) {
                if (*nsiz==0) *nsiz=1024; else *nsiz+=1024;
                if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*(*nsiz)))) {
                    trace(1,"readrnxn: memory allocation error (n=%d)\n",*nsiz);
                    *nsiz=nav->n=0; return -1;
                }
            }
            eph=&nav->eph[nav->n++];
            eph->sat=sat; eph->iode=(int)e[3]; eph->iodc=(int)e[26];
            eph->sva=(int)e[23]; eph->svh=(int)e[24];
            eph->toc=toc; 
            eph->toe=adjweek(gpst2time((int)e[21],e[11]),toc);
            eph->ttr=adjweek(gpst2time((int)e[21],e[27]),toc);
            eph->A=e[10]*e[10]; eph->e=e[8]; eph->i0  =e[15]; eph->OMG0=e[13];
            eph->omg =e[17]; eph->M0 =e[ 6]; eph->deln=e[ 5]; eph->OMGd=e[18];
            eph->idot=e[19]; eph->crc=e[16]; eph->crs =e[ 4]; eph->cuc =e[ 7];
            eph->cus =e[ 9]; eph->cic=e[12]; eph->cis =e[14]; eph->toes=e[11];
            eph->f0  =e[ 0]; eph->f1 =e[ 1]; eph->f2  =e[ 2]; eph->tgd =e[25];
            eph->fit =e[28];
            i=0;
        }
    }
    trace(3,"readrnxn: nnav=%d\n",nav->n);
    return 0;
}
/* read rinex file -----------------------------------------------------------*/
static char readrnxfp(FILE *fp, gtime_t ts, gtime_t te, double tint, int rcv, 
                      obs_t *obs, nav_t *nav, int *osiz, int *nsiz)
{
    double ion[8]={ /* default ionosphere model parameters */
        0.1676E-7,-0.7451E-8,-0.5960E-7,0.1192E-6,1.352E5,-1.802E5,6.554E4,0.0
    };
    double utc[4]={0};
    int i,nt,leaps=0;
    char type,tobs[32][3];
    
    trace(3,"readrnxfp: osiz=%d nsiz=%d\n",*osiz,*nsiz);
    
    type=readrnxh(fp,tobs,&nt,ion,utc,&leaps);
    switch (type) {
    case 'O':
        if (obs) {
            if (readrnxo(fp,ts,te,tint,tobs,nt,rcv,obs,osiz)<0) return 'E';
        }
        break;
    case 'N':
        if (nav) {
            if (readrnxn(fp,nav,nsiz)<0) return 'E';
            for (i=0;i<8;i++) nav->ion[i]=ion[i];
            for (i=0;i<4;i++) nav->utc[i]=utc[i];
            nav->leaps=leaps;
        }
        break;
    default :
        trace(1,"readrnxfp: unknown rinex type (type=%c)\n",type);
        break;
    }
    return type;
}
/* read rinex file ------------------------------------------------------------
* read rinex obs/nav file and add them obs/nav struct
* args   : char *file    I      rinex file
*          gtime_t ts,te I      observation time start/end (ts>=te:all)
*          double tint   I      observation time interval (sec) (0:all)
*          int   rcv     I      receiver number
*          obs_t *obs    IO     observation data    (NULL: no input)
*          nav_t *nav    IO     navigation messages (NULL: no input)
*          int   *osiz   IO     max observation data size
*          int   *nsiz   IO     max navigation message size
* return : status ('O':obs read,'N':nav read,others:error)
*-----------------------------------------------------------------------------*/
extern char readrnxf(const char *file, gtime_t ts, gtime_t te, double tint,
                     int rcv, obs_t *obs, nav_t *nav, int *osiz, int *nsiz)
{
    FILE *fp;
    int cstat;
    char type,tmpfile[1024],s1[64],s2[64];
    
    time2str(ts,s1,0); time2str(te,s2,0);
    trace(3,"readrnxf: file=%s ts=%s te=%s tint=%.1f rcv=%d\n",file,s1,s2,tint,rcv);
    
    if ((cstat=uncompress(file,tmpfile))<0) {
        trace(1,"readrnxf: file uncompact error (%s)\n",file);
        return 'E';
    }
    if (!(fp=fopen(cstat?tmpfile:file,"r"))) {
        trace(1,"readrnxf: file open error (%s)\n",cstat?tmpfile:file);
        return 'E';
    }
    type=readrnxfp(fp,ts,te,tint,rcv,obs,nav,osiz,nsiz);
    fclose(fp);
    if (cstat) remove(tmpfile); /* delete temporary file */
    return type;
}
/* read rinex files in time range/interval -------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char *files[] I      rinex files (wild-card * expanded)
*          int   n       I      number of rinex files (0:input from stdin)
*          gtime_t ts,te I      observation time start/end (ts>=te:all)
*          double tint   I      observation time interval (sec) (0:all)
*          obs_t *obs    IO     observation data    (NULL: no input)
*          nav_t *nav    IO     navigation messages (NULL: no input)
* return : number of epochs (0:no data,<0:error)
*-----------------------------------------------------------------------------*/
extern int readrnxt(char *files[], int n, gtime_t ts, gtime_t te, double tint,
                    obs_t *obs, nav_t *nav)
{
    int i,j,m,rcv=1,nobs=0,osiz=0,nsiz=0;
    char type,*efiles[MAXFILES]={0},s1[64],s2[64];
    
    time2str(ts,s1,0); time2str(te,s2,0);
    trace(3,"readrnxt: file=%s n=%d ts=%s te=%s tint=%.1f\n",n>0?files[0]:"",n,
          s1,s2,tint);
    
    if (obs) {obs->n=0; obs->data=NULL;}
    if (nav) {nav->n=0; nav->eph=NULL;}
    if (n>0) {
        for (i=0;i<MAXFILES;i++) {
            if (!(efiles[i]=(char *)malloc(1024))) {
                for (i--;i>=0;i--) free(efiles[i]); return -1;
            }
        }
        for (i=0;i<n;i++) {
            m=expath(files[i],efiles,MAXFILES);
            for (j=0,type=' ';j<m;j++) {
                type=readrnxf(efiles[j],ts,te,tint,rcv,obs,nav,&osiz,&nsiz);
                if (type=='E') continue;
            }
            if (type=='O') rcv++;
        }
        for (i=0;i<MAXFILES;i++) free(efiles[i]);
    }
    else readrnxfp(stdin,ts,te,tint,rcv,obs,nav,&osiz,&nsiz);
    
    if (obs&&obs->n>0) {
        if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*obs->n))) {
            obs->n=nav->n=0; free(nav->eph);
            return -1;
        }
        nobs=sortobs(obs);
    }
    if (nav&&nav->n>0) {
        uniqeph(nav);
        if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*nav->n))) {
            obs->n=nav->n=0; free(obs->data);
            return -1;
        };
    }
    return nobs;
}
/* read rinex files ------------------------------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char *files[] I      rinex files (wild-card * expanded)
*          int   n       I      number of rinex files (0:input from stdin)
*          obs_t *obs    O      observation data
*          nav_t *nav    O      navigation messages
* return : number of epochs (0:no data,<0:error)
* note   : obs data are sorted by time, receiver and satellite number
*          RINEX 2.10, only support gps.
*-----------------------------------------------------------------------------*/
extern int readrnx(char *files[], int n, obs_t *obs, nav_t *nav)
{
    gtime_t t={0};
    
    trace(3,"readrnx: file=%s n=%d\n",n>0?files[0]:"",n);
    
    return readrnxt(files,n,t,t,0.0,obs,nav);
}
/* output rinex obs header -----------------------------------------------------
* output rinex obd file header
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          char   tobs[][3] I   observation types
*          int    ntobs     I   number of observation types
*          gtime_t time     I   time of first observation record
* return : none
*-----------------------------------------------------------------------------*/
extern void outrnxobsh(FILE *fp, const rnxopt_t *opt, char tobs[][3], int ntobs,
                       gtime_t time)
{
    double ep[6],pos[3]={0},del[3]={0};
    int i;
    char date[32];
    
    trace(3,"outrnxobsh: prog=%s runby=%s comment=%s\n",opt->prog,opt->runby,
          opt->comment[0]);
    
    time2str(timeget(),date,0);
    time2epoch(time,ep);
    fprintf(fp,"%9.2f%-11s%-20s%-20s%-20s\n",RNXVER,"","OBSERVATION DATA",
            "G (GPS)","RINEX VERSION / TYPE");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->prog,opt->runby,date,
            "PGM / RUN BY / DATE");
    for (i=0;i<MAXCOMMENT;i++) {
        if (*opt->comment[i]) fprintf(fp,"%-60.60s%-20s\n",opt->comment[i],"COMMENT");
    }
    fprintf(fp,"%-60.60s%-20s\n",opt->marker,"MARKER NAME");
    fprintf(fp,"%-20.20s%-40.40s%-20s\n",opt->name[0],opt->name[1],"OBSERVER / AGENCY");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->rec[0],opt->rec[1],
            opt->rec[2],"REC # / TYPE / VERS");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->ant[0],opt->ant[1],
            opt->ant[2],"ANT # / TYPE");
    for (i=0;i<3;i++) if (fabs(opt->apppos[i])<1E8) pos[i]=opt->apppos[i];
    for (i=0;i<3;i++) if (fabs(opt->antdel[i])<1E8) del[i]=opt->antdel[i];
    fprintf(fp,"%14.4f%14.4f%14.4f%-18s%-20s\n",pos[0],pos[1],pos[2],"",
            "APPROX POSITION XYZ");
    fprintf(fp,"%14.4f%14.4f%14.4f%-18s%-20s\n",del[0],del[1],del[2],"",
            "ANTENNA: DELTA H/E/N");
    fprintf(fp,"%6d%6d%-48s%-20s\n",1,1,"","WAVELENGTH FACT L1/2");
    for (i=0;i<ntobs;i++) {
        if (i==0) fprintf(fp,"%6d",ntobs); else if (i%9==0) fprintf(fp,"      ");
        fprintf(fp,"%6s",tobs[i]);
        if (i%9==8||i==ntobs-1) {
            fprintf(fp,"%*s%-20s\n",6*(i%9-8),"","# / TYPES OF OBSERV");
        }
    }
    fprintf(fp,"  %04.0f%6.0f%6.0f%6.0f%6.0f%13.7f     %-12s%-20s\n",ep[0],
            ep[1],ep[2],ep[3],ep[4],ep[5],"GPS","TIME OF FIRST OBS");
    fprintf(fp,"%-60.60s%-20s\n","","END OF HEADER");
}
/* output obs data field -----------------------------------------------------*/
static void outrnxobsf(FILE *fp, double obs, int lli)
{
    if (obs==0.0||obs<=-1E9||obs>=1E9) fprintf(fp,"              ");
    else fprintf(fp,"%14.3f",obs);
    if (lli<=0) fprintf(fp,"  "); else fprintf(fp,"%1.1d ",lli);
}
/* output rinex obs body -------------------------------------------------------
* output rinex obs body
* args   : FILE   *fp       I   output file pointer
*          obsd_t *obs      I   observation data
*          int    n         I   number of observation data
*          int    epflag    I   epoch flag (0:ok,1:power failure,>1:event flag)
*          char   tobs[][3] I   observation types
*          int    ntobs     I   number of observation types
* return : none
*-----------------------------------------------------------------------------*/
extern void outrnxobsb(FILE *fp, const obsd_t *obs, int n, int epflag,
                       char tobs[][3], int ntobs)
{
    double ep[6];
    int i,j;
    
    time2epoch(obs[0].time,ep);
    fprintf(fp," %02d %2.0f %2.0f %2.0f %2.0f%11.7f  %d%3d",
            (int)ep[0]%100,ep[1],ep[2],ep[3],ep[4],ep[5],epflag,n);
    for (i=0;i<n;i++) {
        if (i>0&&i%12==0) fprintf(fp,"\n%32s","");
        if      (obs[i].sat< 33) fprintf(fp,"G%2d",obs[i].sat);
        else if (obs[i].sat<100) fprintf(fp,"R%2d",obs[i].sat-37);
        else                     fprintf(fp,"S%2d",obs[i].sat-100);
    }
    fprintf(fp,"\n");
    for (i=0;i<n;i++) {
        for (j=0;j<ntobs;j++) {
            if (tobs[j][0]=='L') {
                outrnxobsf(fp,obs[i].L[tobs[j][1]-'1'],obs[i].LLI[tobs[j][1]-'1']);
            }
            else if (tobs[j][0]=='P'||tobs[j][0]=='C') {
                outrnxobsf(fp,obs[i].P[tobs[j][1]-'1'],-1);
            }
            else if (tobs[j][0]=='D') {
                outrnxobsf(fp,obs[i].D[tobs[j][1]-'1'],-1);
            }
            else if (tobs[j][0]=='S') {
                outrnxobsf(fp,obs[i].SNR[tobs[j][1]-'1'],-1);
            }
            else fprintf(fp,"               ");
            if (j==ntobs-1||j%5==4) fprintf(fp,"\n");
        }
    }
}
/* output nav member by rinex nav format -------------------------------------*/
static void outnavf(FILE *fp, double value)
{
    double e=fabs(value)<1E-99?0.0:floor(log10(fabs(value))+1.0);
    fprintf(fp," %s.%012.0fE%+03.0f",value<0.0?"-":" ",fabs(value)/pow(10.0,e-12.0),e);
}
/* output rinex nav header -----------------------------------------------------
* output rinex nav file header
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          double *ion      I   ionospheric parameters (NULL: no output)
*          double *dutc     I   delta-utc parameters (NULL: no output)
*          int    leaps     I   leap seconds (s) (<0: no output)
* return : none
*-----------------------------------------------------------------------------*/
extern void outrnxnavh(FILE *fp, const rnxopt_t *opt, const double *ion,
                       const double *dutc, int leaps)
{
    int i;
    char date[64];
    
    trace(3,"outrnxnavh: prog=%s runby=%s comment=%s\n",opt->prog,opt->runby,
          opt->comment[0]);
    
    time2str(timeget(),date,0);
    fprintf(fp,"%9.2f           %-20s%-20s%-20s\n",RNXVER,"N: GPS NAV DATA","",
            "RINEX VERSION / TYPE");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->prog,opt->runby,date,
            "PGM / RUN BY / DATE");
    for (i=0;i<MAXCOMMENT;i++) {
        if (*opt->comment[i]) fprintf(fp,"%-60.60s%-20s\n",opt->comment[i],"COMMENT");
    }
    if (ion) {
        fprintf(fp,"  %12.4E%12.4E%12.4E%12.4E%10s%-20s\n",ion[0],ion[1],ion[2],
                ion[3],"","ION ALPHA");
        fprintf(fp,"  %12.4E%12.4E%12.4E%12.4E%10s%-20s\n",ion[4],ion[5],ion[6],
                ion[7],"","ION BETA");
    }
    if (dutc) {
        fprintf(fp,"   "); outnavf(fp,dutc[0]); outnavf(fp,dutc[1]);
        fprintf(fp,"%9.0f%9.0f %-20s\n",dutc[2],dutc[3],"DELTA-UTC: A0,A1,T,W");
    }
    if (leaps>=0) {
        fprintf(fp,"%6d%54s%-20s\n",leaps,"","LEAP SECONDS");
    }
    fprintf(fp,"%60s%-20s\n","","END OF HEADER");
}
/* output rinex nav body -------------------------------------------------------
* output rinex nav file body record
* args   : FILE   *fp       I   output file pointer
*          eph_t  *eph      I   ephemeris record
* return : none
*-----------------------------------------------------------------------------*/
extern void outrnxnavb(FILE *fp, const eph_t *eph)
{
    double t[6],ttr;
    int week,w;
    
    (void)time2gpst(eph->toe,&week);
    time2epoch(eph->toc,t);
    fprintf(fp,"%2d %02d %2.0f %2.0f %2.0f %2.0f %4.1f",eph->sat,(int)t[0]%100,
            t[1],t[2],t[3],t[4],t[5]);
    outnavf(fp,eph->f0  );
    outnavf(fp,eph->f1  );
    outnavf(fp,eph->f2  ); fprintf(fp,"\n   ");
    outnavf(fp,eph->iode);
    outnavf(fp,eph->crs );
    outnavf(fp,eph->deln);
    outnavf(fp,eph->M0  ); fprintf(fp,"\n   ");
    outnavf(fp,eph->cuc );
    outnavf(fp,eph->e   );
    outnavf(fp,eph->cus );
    outnavf(fp,sqrt(eph->A)); fprintf(fp,"\n   ");
    outnavf(fp,eph->toes);
    outnavf(fp,eph->cic );
    outnavf(fp,eph->OMG0);
    outnavf(fp,eph->cis ); fprintf(fp,"\n   ");
    outnavf(fp,eph->i0  );
    outnavf(fp,eph->crc );
    outnavf(fp,eph->omg );
    outnavf(fp,eph->OMGd); fprintf(fp,"\n   ");
    outnavf(fp,eph->idot);
    outnavf(fp,1);
    outnavf(fp,week);
    outnavf(fp,0); fprintf(fp,"\n   ");
    outnavf(fp,eph->sva );
    outnavf(fp,eph->svh );
    outnavf(fp,eph->tgd );
    outnavf(fp,eph->iodc); fprintf(fp,"\n   ");
    ttr=time2gpst(eph->ttr,&w);
    outnavf(fp,ttr+(w-week)*604800.0);
    outnavf(fp,eph->fit ); fprintf(fp,"\n");
}
/* uncompress file -------------------------------------------------------------
* uncompress (uncompress/unzip/uncompact hatanaka-compression) file
* args   : char   *file     I   input file
*          char   *uncfile  O   uncompressed file
* return : status (-1:error,0:not compressed file,1:uncompress completed)
* note   : creates uncompressed file in tempolary directory
*          gzip and crx2rnx command have to be installed in commands path
*-----------------------------------------------------------------------------*/
extern int uncompress(const char *file, char *uncfile)
{
    int stat=0;
    char *p,cmd[1024]="",tmpfile[1024]="";
    
    trace(3,"uncompress: file=%s\n",file);
    
    strcpy(tmpfile,file);
    
    /* uncompress by gzip */
    if (!(p=strrchr(tmpfile,'.'))) return 0;
    if (!strcmp(p,".z")||!strcmp(p,".Z")||!strcmp(p,".gz")||!strcmp(p,".GZ")||
        !strcmp(p,".zip")||!strcmp(p,".ZIP")) {
        strcpy(uncfile,tmpfile); uncfile[p-tmpfile]='\0';
        sprintf(cmd,"gzip -f -d -c \"%s\" > \"%s\"",tmpfile,uncfile);
        if (execcmd(cmd)) {
            remove(uncfile);
            return -1;
        }
        strcpy(tmpfile,uncfile);
        stat=1;
    }
    /* extract hatanaka-compressed file by cnx2rnx */
    if ((p=strrchr(tmpfile,'.'))&&strlen(p)>3&&(*(p+3)=='d'||*(p+3)=='D')) {
        strcpy(uncfile,tmpfile); uncfile[p-tmpfile+3]=*(p+3)=='D'?'O':'o';
        sprintf(cmd,"crx2rnx < \"%s\" > \"%s\"",tmpfile,uncfile);
        if (execcmd(cmd)) {
            remove(uncfile);
            if (stat) remove(tmpfile);
            return -1;
        }
        if (stat) remove(tmpfile);
        stat=1;
    }
    trace(3,"uncompress: stat=%d\n",stat);
    return stat;
}
