/*------------------------------------------------------------------------------
* solution.c : solution functions
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/11/03  1.0 new
*           2009/01/05  1.1 add function outsols(), outsolheads(),
*                           setsolformat(), outsolexs, outsolex
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: solution.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

#define COMMENTH    "%"         /* comment line indicator */
#define MAXFIELD    32          /* max number of fields in a record */
#define MAXRPOS      8          /* max number of reference position */
#define MAXNMEA    256          /* max length of nmea sentence */
#define MAXSOLH    4096         /* max length of solution header */
#define MAXSOL     1024         /* max length of solution body */
#define KNOT2M     0.514444444  /* m/knot */
#define SQRT(x)    ((x)<0.0?0.0:sqrt(x))

/* global variables ----------------------------------------------------------*/
static solopt_t opt={           /* solution options */
    0,0,0,3,0,1,0,0,0,0," "
};
/* separate fields -----------------------------------------------------------*/
static int tonum(char *p, const char *sep, double v[])
{
    int n=0,m=(int)strlen(sep);
    char *q;
    
    if (!strcmp(sep,"")||!strcmp(sep," ")) {
        for (p=strtok(p,sep);p&&n<MAXFIELD;p=strtok(NULL,sep)) v[n++]=atof(p);
        return n;
    }
    while (n<MAXFIELD) {
        if ((q=strstr(p,sep))) *q='\0'; 
        v[n++]=atof(p);
        if (q) p=q+m; else break;
    }
    return n;
}
/* sqrt of covariance --------------------------------------------------------*/
static double sqvar(double covar)
{
    return covar<0.0?-sqrt(-covar):sqrt(covar);
}
/* convert dd/mm/ss to deg in nmea format ------------------------------------*/
static double dms2deg(const double dms[])
{
    return dms[0]+dms[1]/60.0+dms[2]/3600.0;
}
/* convert ddmm.mm in nmea format to deg -------------------------------------*/
static double dmm2deg(double dmm)
{
    return floor(dmm/100.0)+fmod(dmm,100.0)/60.0;
}
/* convert time in nmea format to time ---------------------------------------*/
static void septime(double t, double *t1, double *t2, double *t3)
{
    *t1=floor(t/10000.0); t-=*t1*10000.0; *t2=floor(t/100.0); *t3=t-*t2*100.0;
}
/* convert date/time to time struct ------------------------------------------*/
static gtime_t datet2time(double date, double time)
{
    double ep[6];
    
    septime(date,ep+2,ep+1,ep);
    septime(time,ep+3,ep+4,ep+5);
    ep[0]+=ep[0]<80.0?2000.0:1900.0;
    return epoch2time(ep);
}
/* convert deg to deg/min/sec ------------------------------------------------*/
static void degtodms(double deg, double *dms)
{
    double sgn=1.0;
    
    if (deg<0.0) {deg=-deg; sgn=-1.0;}
    dms[0]=sgn*floor(deg);
    dms[1]=floor((deg-dms[0])*60.0);
    dms[2]=(deg-dms[0]-dms[1]/60.0)*3600;
}
/* screen and set solution data ----------------------------------------------*/
static void screensol(gtime_t ts, gtime_t te, double tint, int qflag,
                      solp_t *solp, int *nsol, gtime_t time, double *pos,
                      int posq, float *poss)
{
    int i;
    
    if (ts.time!=0&&timediff(time,ts)<0.0) return;
    if (te.time!=0&&timediff(time,te)>0.0) return;
    if (tint>0.0) {
        if (fmod(time2gpst(time,NULL)+DTTOL,tint)>2.0*DTTOL) return;
    }
    if (qflag>0&&posq!=qflag) return;
    
    if (solp->n>=*nsol) {
        *nsol=*nsol==0?8192:*nsol*2;
        if (!(solp->data=(solpd_t *)realloc(solp->data,sizeof(solpd_t)*(*nsol)))) {
            fprintf(stderr,"%s : memory allocation error\n",__FILE__);
            *nsol=solp->n=0; return;
        }
    }
    if (solp->n>0&&timediff(solp->data[solp->n-1].time,time)==0.0&&pos[3]==0.0)
        return;
    solp->data[solp->n].time=time;
    for (i=0;i<3;i++) solp->data[solp->n].pos[i]=pos[i];
    for (i=0;i<6;i++) solp->data[solp->n].poss[i]=poss[i];
    solp->data[solp->n++].posq=(short)posq;
}
/* decode nmea sentence -----------------------------------------------------*/
static int decodenmea(char *sent, gtime_t *time, double *sol, int *solq,
                      float *sols, double *date)
{
    const int qs[]={0,5,4,0,2,1,0,0,0,4};
    double tod=0.0,lat=0.0,lon=0.0,vel,dir,hdop,alt=0.0,msl=0.0,ang=0.0;
    int n,qflg=5,nrcv,type=0;
    char *p,*q,val=' ',ns='N',ew='E',mew='E',mode='N',ua,um;
    
    if (!strncmp(sent,"$GPRMC",6)) {
        for (p=sent+7,n=0;p;p=q+1,n++) {
            if ((q=strchr(p,','))) *q='\0';
            else if ((q=strchr(p,'*'))) *q='\0';
            else break;
            switch (n) {
            case  0: tod =atof(p); break;
            case  1: val =*p;      break;
            case  2: lat =atof(p); break;
            case  3: ns  =*p;      break;
            case  4: lon =atof(p); break;
            case  5: ew  =*p;      break;
            case  6: vel =atof(p); break;
            case  7: dir =atof(p); break;
            case  8: *date=atof(p); break;
            case  9: ang=atof(p);  break;
            case 10: mew=*p;       break;
            case 11: mode=*p;      break;
            }
        }
        if (n<=8||(val!='A'&&val!='D')||(mode!='A'&&mode!='D')) return 0;
        alt=msl=0.0;
        *solq=mode=='D'?4:5;
        type=1;
    }
    else if (!strncmp(sent,"$GPGGA",6)) {
        for (p=sent+7,n=0;p;p=q+1,n++) {
            if ((q=strchr(p,','))) *q='\0';
            else if ((q=strchr(p,'*'))) *q='\0';
            else break;
            switch (n) {
            case  0: tod =atof(p); break;
            case  1: lat =atof(p); break;
            case  2: ns  =*p;      break;
            case  3: lon =atof(p); break;
            case  4: ew  =*p;      break;
            case  5: qflg=atoi(p); break;
            case  6: nrcv=atoi(p); break;
            case  7: hdop=atof(p); break;
            case  8: alt =atof(p); break;
            case  9: ua  =*p;      break;
            case 10: msl =atof(p); break;
            case 11: um  =*p;      break;
            }
        }
        if (*date==0.0||n<=11||(ns!='N'&&ns!='S')||(ew!='E'&&ns!='W')) return 0;
        *solq=qs[qflg<0||9<qflg?0:qflg];
        type=2;
    }
    else return 0;
    
    *time=utc2gpst(datet2time(*date,tod));
    sol[0]=(ns=='N'?1.0:-1.0)*dmm2deg(lat);
    sol[1]=(ew=='E'?1.0:-1.0)*dmm2deg(lon);
    sol[2]=alt+msl;
    sols[0]=sols[1]=sols[2]=0.0f;
    return type;
}
/* read nmea file -----------------------------------------------------------*/
static int readnmea(FILE *fp, gtime_t ts, gtime_t te, double tint, int qflag,
                    solp_t *solp, int *nsol)
{
    gtime_t time,timep={0};
    double pos[3],date=0.0,t;
    float poss[6];
    int i,j=-1,nb,posq,res=1,type;
    char buff[2048],sent[MAXNMEA+1];
    
    while ((nb=(int)fread(buff,1,sizeof(buff),fp))>0) {
        for (i=0;i<nb;i++) {
            if (buff[i]=='$') j=0;
            if (j<0) continue;
            if ((sent[j++]=buff[i])=='\n') {
                sent[j]='\0';
                type=decodenmea(sent,&time,pos,&posq,poss,&date);
                if (type) {
                    if (timep.time!=0) {
                        t=timediff(time,timep);
                        if      (t<-43200.0) time=timeadd(time, 86400.0);
                        else if (t> 43200.0) time=timeadd(time,-86400.0);
                    }
                    timep=time;
                    if (type==2) {
                        screensol(ts,te,tint,qflag,solp,nsol,time,pos,posq,poss);
                    }
                    res=0;
                }
                j=-1;
            }
            if ((j==3&&strncmp(sent,"$GP",3))||j>=MAXNMEA) j=-1;
        }
    }
    return res;
}
/* decode position record ----------------------------------------------------*/
static int decodepos(char *buff, int *posf, int *degf, int *tsys, char *sep,
                     gtime_t *time, double *sol, int *solq, float *sols,
                     char *rpos, int *week, char *inputs[], int *ninp)
{
    double v[MAXFIELD],pos[3],P[9]={0},Q[9];
    int i,m;
    char *p=buff;
    
    if (*buff=='%'||*buff=='#'||*buff==';') { /* comment line */
        if (strstr(buff,"ref pos")||strstr(buff,"slave pos")) {
            if (!(p=strchr(buff,':'))) return -1;
            strcpy(rpos,p+1);
            return 1;
        }
        else if (strstr(buff,"inputs")&&inputs&&ninp) {
            for (p=strtok(p+1," ");p;p=strtok(NULL," "))
                strcpy(inputs[(*ninp)++],p);
        }
        if      (strstr(buff,"GPST")) *tsys=TIMES_GPST;
        else if (strstr(buff,"UTC" )) *tsys=TIMES_UTC;
        else if (strstr(buff,"JST" )) *tsys=TIMES_JST;
        if ((p=strstr(buff,"x-ecef(m)"))) {
            *posf=SOLF_XYZ; *degf=0; strncpy(sep,p+9,1); sep[1]='\0';
        }
        else if ((p=strstr(buff,"e-baseline(m)"))) {
            *posf=SOLF_ENU; *degf=0; strncpy(sep,p+13,1); sep[1]='\0';
        }
        else if ((p=strstr(buff,"latitude(d'\")"))) {
            *posf=SOLF_LLH; *degf=1; strncpy(sep,p+14,1); sep[1]='\0';
        }
        else if ((p=strstr(buff,"latitude(deg)"))) {
            *posf=SOLF_LLH; *degf=0; strncpy(sep,p+13,1); sep[1]='\0';
        }
        return -1;
    }
    if (sscanf(buff,"%lf/%lf/%lf %lf:%lf:%lf",v,v+1,v+2,v+3,v+4,v+5)>=6) {
        *time=epoch2time(v);
        if (*tsys==1) *time=utc2gpst(*time);
        else if (*tsys==2) *time=utc2gpst(timeadd(*time,-9*3600.0));
        m=tonum(buff+11,sep,v);
    }
    else if ((m=tonum(buff,sep,v))>=2) {
        *time=gpst2time((int)v[0],v[1]);
        for (i=0,m--;i<m;i++) v[i]=v[i+1];
    }
    else return -1;
    
    sols[0]=sols[1]=sols[2]=sols[3]=sols[4]=sols[5]=0.0f;
    
    if (*posf==SOLF_XYZ&&m>=4) { /* x/y/z-ecef */
        ecef2pos(v+1,pos);
        sol[0]=pos[0]*R2D; sol[1]=pos[1]*R2D; sol[2]=pos[2];
        *solq=m>=5?(int)v[4]:0;
        if (m>=9) {
            P[0]=v[6]*v[6]; P[4]=v[7]*v[7]; P[8]=v[8]*v[8];
            if (m>=12) {
                P[1]=P[3]=(v[9]<0.0?-1.0:1.0)*v[9]*v[9];
                P[5]=P[7]=(v[10]<0.0?-1.0:1.0)*v[10]*v[10];
                P[2]=P[6]=(v[11]<0.0?-1.0:1.0)*v[11]*v[11];
            }
            covenu(pos,P,Q);
            sols[0]=(float)Q[4]; sols[1]=(float)Q[0]; sols[2]=(float)Q[8];
            sols[3]=(float)sqvar(Q[1]);
            sols[4]=(float)sqvar(Q[2]);
            sols[5]=(float)sqvar(Q[5]);
        }
    }
    else if (*posf==SOLF_ENU&&m>=4) { /* e/n/u-base */
        sol[0]=v[1]; sol[1]=v[2]; sol[2]=v[3];
        *solq=m>=5?(int)v[4]:0;
        if (m>=9) {
            P[0]=v[6]*v[6]; P[4]=v[7]*v[7]; P[8]=v[8]*v[8];
            if (m>=12) {
                P[1]=P[3]=(v[9]<0.0?-1.0:1.0)*v[9]*v[9];
                P[5]=P[7]=(v[10]<0.0?-1.0:1.0)*v[10]*v[10];
                P[2]=P[6]=(v[11]<0.0?-1.0:1.0)*v[11]*v[11];
            }
            sols[0]=(float)P[0]; sols[1]=(float)P[4]; sols[2]=(float)P[8];
            sols[3]=(float)sqvar(P[1]);
            sols[4]=(float)sqvar(P[5]);
            sols[5]=(float)sqvar(P[2]);
        }
    }
    else if (*degf&&m>=8) { /* lat/lon/height (ddd mm ss) */
        sol[0]=dms2deg(v+1); sol[1]=dms2deg(v+4); sol[2]=v[7];
        *solq=m>=9?(int)v[8]:0;
        if      (m>=16) for (i=0;i<6;i++) sols[i]=(float)v[i+10];
        else if (m>=13) for (i=0;i<3;i++) sols[i]=(float)v[i+10];
    }
    else if (m>=4) { /* lat/lon/height (ddd.dddd) */
        sol[0]=v[1]; sol[1]=v[2]; sol[2]=v[3];
        *solq=m>=5?(int)v[4]:0;
        if      (m>=12) for (i=0;i<6;i++) sols[i]=(float)v[i+6];
        else if (m>= 9) for (i=0;i<3;i++) sols[i]=(float)v[i+6];
    }
    if (*solq<0||6<*solq) *solq=0;
    return m>=4?0:-1;
}
/* reference position -------------------------------------------------------*/
static void torpos(char str[][256], int n, int posf, int degf, char *sep,
                   double rpos[][3])
{
    double v[MAXFIELD];
    int i,j,m;
    
    for (i=0;i<n;i++) {
        m=tonum(str[i],sep,v);
        if (posf==SOLF_LLH&&degf==0&&m>=3) {
            for (j=0;j<3;j++) rpos[i][j]=v[j];
        }
        else if (posf==SOLF_LLH&&degf==1&&m>=7) {
            rpos[i][0]=dms2deg(v);
            rpos[i][1]=dms2deg(v+3); rpos[i][2]=v[6];
        }
        else if (posf==SOLF_XYZ&&m>=3) {
            ecef2pos(v,rpos[i]);
            rpos[i][0]*=R2D; rpos[i][1]*=R2D;
        }
        else for (j=0;j<3;j++) rpos[i][j]=0.0;
    }
}
/* read position file -------------------------------------------------------*/
static int readsolf(FILE *fp, gtime_t ts, gtime_t te, double tint, int qflag,
                    solp_t *solp, int *nsol, double rpos[][3], int *nrpos,
                    char *inputs[], int *ninp)
{
    gtime_t time;
    double pos[3];
    float poss[6];
    int week=1400,posf=-1,degf=0,tsys=0,stat,posq;
    char buff[MAXSOL],sep[16]=" ",str[MAXRPOS][256];
    
    *nrpos=0;
    while (fgets(buff,sizeof(buff),fp)) {
        stat=decodepos(buff,&posf,&degf,&tsys,sep,&time,pos,&posq,poss,
                       str[*nrpos],&week,inputs,ninp);
        if (stat==0) {
            screensol(ts,te,tint,qflag,solp,nsol,time,pos,posq,poss);
        }
        else if (stat==1&&*nrpos<MAXRPOS) (*nrpos)++;
    }
    torpos(str,*nrpos,posf,degf,sep,rpos);
    return posf==1?0:posf;
}
/* compare solution data ----------------------------------------------------*/
static int cmpsolp(const void *p1, const void *p2)
{
    solpd_t *q1=(solpd_t *)p1,*q2=(solpd_t *)p2;
    double tt=timediff(q1->time,q2->time);
    return tt<-0.0?-1:(tt>0.0?1:0);
}
/* read screened solutions -----------------------------------------------------
* read solution from file as text position format screened by time span, time
* interval and quality flag
* args   : char   *files[]  I  solution files
*          int    nfile     I  number of solution files
*          gtime_t ts       I  start time 
*          gtime_t te       I  end time 
*          double tint      I  time interval 
*          int    qflag     I  quality flag 
*          solp_t *solp     O  position solutions
*          double rpos[][]  IO ref station pos {lat,lon,height} (NULL: no input)
*          int    *nrpos    IO number of ref station pos (NULL: no input)
*          char   **inputs  IO input files (NULL: no input)
*          int    *ninp     IO number of input files (NULL: no input)
* return : status(-1:no data,0:lat/lon/hgt,1:x/y/z-ecef,2:e/n/u-baseline)
*-----------------------------------------------------------------------------*/
extern int readsolt(char *files[], int nfile, gtime_t ts, gtime_t te,
                    double tint, int qflag, solp_t *solp, double rpos[][3],
                    int *nrpos, char **inputs, int *ninp)
{
    FILE *fp;
    int i,nsol=0,posf=-1;
    char s1[64],s2[64];
    
    time2str(ts,s1,3); time2str(te,s2,3);
    trace(3,"readsolt:file=%s nfile=%d ts=%s te=%s tint=%.1f qflag=%d\n",
          nfile>0?files[0]:"",nfile,s1,s2,tint,qflag);
    
    if (nrpos) *nrpos=0;
    if (ninp) *ninp=0;
    for (i=0;i<nfile;i++) {
        if (!(fp=fopen(files[i],"rb"))) continue;
        if (readnmea(fp,ts,te,tint,qflag,solp,&nsol)) {
            rewind(fp);
            posf=readsolf(fp,ts,te,tint,qflag,solp,&nsol,rpos,nrpos,inputs,ninp);
        }
        else posf=SOLF_LLH;
        fclose(fp);
    }
    if (posf<0) {
        if (solp->n>0) free(solp->data);
        solp->n=0; solp->data=NULL;
        return -1;
    }
    if (!(solp->data=(solpd_t *)realloc(solp->data,sizeof(solpd_t)*solp->n))) {
        fprintf(stderr,"%s : memory allocation error\n",__FILE__);
        solp->n=0; return 0;
    }
    if (solp->n>0) qsort(solp->data,solp->n,sizeof(solpd_t),cmpsolp);
    return posf;
}
/* read solutions -------------------------------------------------------------
* read position solution from file as text position format
* args   : char   *files[]  I  solution files
*          int    nfile     I  number of solution files
*          solp_t *solp     O  position solutions
*          double rpos[][]  IO ref station pos {lat,lon,height} (NULL: no input)
*          int    *nrpos    IO number of ref station pos (NULL: no input)
*          char   **inputs  IO input files (NULL: no input)
*          int    *ninp     IO number of input files (NULL: no input)
* return : number of solution data
*-----------------------------------------------------------------------------*/
extern int readsol(char *files[], int nfile, solp_t *solp, double rpos[][3],
                   int *nrpos, char **inputs, int *ninp)
{
    gtime_t t={0};
    
    trace(3,"readsol:file=%s nfile=%d\n",nfile>0?files[0]:"",nfile);
    
    return readsolt(files,nfile,t,t,0.0,0,solp,rpos,nrpos,inputs,ninp);
}
/* output solution as the form of x/y/z-ecef ---------------------------------*/
static int outecef(unsigned char *buff, const char *s, const double *rr,
                   const double *Qr, int stat, int ns)
{
    char *p=(char *)buff,*sep=opt.sep;
    
    trace(3,"outecef:\n");
    
    if (stat<=SOLQ_NONE) return 0;
    
    p+=sprintf(p,"%s%s%14.4f%s%14.4f%s%14.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f\n",
               s,sep,rr[0],sep,rr[1],sep,rr[2],sep,stat,sep,ns,
               sep,SQRT(Qr[0]),sep,SQRT(Qr[4]),sep,SQRT(Qr[8]),
               sep,sqvar(Qr[1]),sep,sqvar(Qr[5]),sep,sqvar(Qr[2]));
    return p-(char *)buff;
}
/* output solution as the form of lat/lon/height -----------------------------*/
static int outpos(unsigned char *buff, const char *s, const double *rr,
                  const double *Qr, int stat, int ns)
{
    double pos[3],dms1[3],dms2[3],Qe[9];
    char *p=(char *)buff,*sep=opt.sep;
    
    trace(3,"outpos:\n");
    
    if (stat<=SOLQ_NONE) return 0;
    
    ecef2pos(rr,pos); covenu(pos,Qr,Qe);
    
    if (opt.height==1) { /* geodetic height */
        pos[2]-=geoidh(pos);
    }
    if (opt.degf) {
        degtodms(pos[0]*R2D,dms1); degtodms(pos[1]*R2D,dms2);
        p+=sprintf(p,"%s%s%4.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f",s,sep,
                   dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],sep,
                   dms2[2]);
    }
    else p+=sprintf(p,"%s%s%14.9f%s%14.9f",s,sep,pos[0]*R2D,sep,pos[1]*R2D);
    p+=sprintf(p,"%s%10.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f\n",
               sep,pos[2],sep,stat,sep,ns,sep,SQRT(Qe[4]),sep,SQRT(Qe[0]),
               sep,SQRT(Qe[8]),sep,sqvar(Qe[1]),sep,sqvar(Qe[2]),sep,sqvar(Qe[5]));
    return p-(char *)buff;
}
/* output solution as the form of e/n/u-baseline -----------------------------*/
static int outenu(unsigned char *buff, const char *s, const double *rr,
                  const double *Qr, const double *rb, int stat, int ns)
{
    double pos[3],r[3],enu[3],Qe[9];
    int i;
    char *p=(char *)buff,*sep=opt.sep;
    
    trace(3,"outenu:\n");
    
    if (stat<=SOLQ_NONE) return 0;
    
    for (i=0;i<3;i++) r[i]=rr[i]-rb[i];
    ecef2pos(rb,pos); covenu(pos,Qr,Qe); ecef2enu(pos,r,enu);
    p+=sprintf(p,"%s%s%14.4f%s%14.4f%s%14.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f%s%8.4f\n",
               s,sep,enu[0],sep,enu[1],sep,enu[2],sep,stat,sep,ns,sep,SQRT(Qe[0]),
               sep,SQRT(Qe[4]),sep,SQRT(Qe[8]),sep,sqvar(Qe[1]),sep,sqvar(Qe[5]),
               sep,sqvar(Qe[1]));
    return p-(char *)buff;
}
/* output solution in the form of nmea RMC sentence --------------------------*/
static int outnmea_rmc(unsigned char *buff, gtime_t t, const double *rr,
                       int stat)
{
    static double dirp=0.0;
    double ep[6],pos[3],enuv[3],dms1[3],dms2[3],vel,dir;
    char *p=(char *)buff,*q,sum;
    
    trace(3,"outnmea_rmc:\n");
    
    if (stat<=SOLQ_NONE) {
        p+=sprintf(p,"$GPRMC,,,,,,,,,,,,");
        for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q;
        p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
        return p-(char *)buff;
    }
    time2epoch(gpst2utc(t),ep);
    ecef2pos(rr,pos);
    ecef2enu(pos,rr+3,enuv);
    vel=norm(enuv,3);
    if (vel>=1.0) {
        dir=atan2(enuv[0],enuv[1])*R2D;
        if (dir<0.0) dir+=360.0;
        dirp=dir;
    }
    else dir=dirp;
    degtodms(fabs(pos[0])*R2D,dms1); degtodms(fabs(pos[1])*R2D,dms2);
    p+=sprintf(p,"$GPRMC,%02.0f%02.0f%05.2f,A,%02.0f%010.7f,%s,%03.0f%010.7f,%s,%4.2f,%4.2f,%02.0f%02.0f%02d,,,%s",
               ep[3],ep[4],ep[5],dms1[0],dms1[1]+dms1[2]/60.0,pos[0]>=0?"N":"S",
               dms2[0],dms2[1]+dms2[2]/60.0,pos[1]>=0?"E":"W",vel/KNOT2M,dir,
               ep[2],ep[1],(int)ep[0]%100,stat<=0?"N":(stat==5?"A":"D"));
    for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q; /* check-sum */
    p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
    return p-(char *)buff;
}
/* output solution in the form of nmea GGA sentence --------------------------*/
static int outnmea_gga(unsigned char *buff, gtime_t t, const double *rr,
                       int stat, int ns)
{
    double h,ep[6],pos[3],dms1[3],dms2[3];
    int qflg[]={0,5,4,0,2,1}; /* quality: 1:sigle,2:dgps,4:float,5:fix,0:other */
    char *p=(char *)buff,*q,sum;
    
    trace(3,"outnmea_gga:\n");
    
    if (stat<=SOLQ_NONE) {
        p+=sprintf(p,"$GPGGA,,,,,,,,,,,,,,");
        for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q;
        p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
        return p-(char *)buff;
    }
    time2epoch(gpst2utc(t),ep);
    ecef2pos(rr,pos);
    h=geoidh(pos);
    degtodms(fabs(pos[0])*R2D,dms1); degtodms(fabs(pos[1])*R2D,dms2);
    p+=sprintf(p,"$GPGGA,%02.0f%02.0f%05.2f,%02.0f%010.7f,%s,%03.0f%010.7f,%s,%d,%02d,,%.3f,M,%.3f,M,,",
               ep[3],ep[4],ep[5],dms1[0],dms1[1]+dms1[2]/60.0,pos[0]>=0?"N":"S",
               dms2[0],dms2[1]+dms2[2]/60.0,pos[1]>=0?"E":"W",qflg[stat],ns,
               pos[2]-h,h);
    for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q; /* check-sum */
    p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
    return p-(char *)buff;
}
/* output solution in the form of nmea GSA sentence --------------------------*/
extern int outnmea_gsa(unsigned char *buff, int stat, int ns, int *sat,
                       double *az, double *el, int *snr, int *vsat)
{
    double azel[MAXSAT*2],dop[4];
    int i,n;
    char *p=(char *)buff,*q,sum;
    
    trace(3,"outnmea_gsa:\n");
    
    if (stat<=SOLQ_NONE) {
        p+=sprintf(p,"$GPGSA,A,1,,,,,,,,,,,,,,,");
        for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q;
        p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
        return p-(char *)buff;
    }
    p+=sprintf(p,"$GPGSA,A,%d",stat<=0?1:3);
    for (i=n=0;i<ns&&n<12;i++) {
        if (!vsat[i]||el[i]<=0.0) continue;
        azel[n*2]=az[i]; azel[1+n*2]=el[i];
        p+=sprintf(p,",%02d",sat[i]);
        n++;
    }
    for (i=n;i<12;i++) p+=sprintf(p,",");
    dops(n,azel,0.0,dop);
    p+=sprintf(p,",%3.1f,%3.1f,%3.1f",dop[1],dop[2],dop[3]);
    for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q; /* check-sum */
    p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
    return p-(char *)buff;
}
/* output solution in the form of nmea GSV sentence --------------------------*/
extern int outnmea_gsv(unsigned char *buff, int stat, int ns, int *sat,
                       double *az, double *el, int *snr, int *vsat)
{
    double azs;
    int i,j,k,n,nmsg,index[MAXOBS];
    char *p=(char *)buff,*q,*s,sum;
    
    trace(3,"outnmea_gsv:\n");
    
    if (stat<=SOLQ_NONE) {
        p+=sprintf(p,"$GPGSV,1,1,0,,,,,,,,,,,,,,,,");
        for (q=(char *)buff+1,sum=0;*q;q++) sum^=*q;
        p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
        return p-(char *)buff;
    }
    for (i=n=0;i<ns&&n<12;i++) {
        if (vsat[i]&&el[i]>0.0) index[n++]=i;
    }
    nmsg=(n-1)/4+1;
    
    for (i=k=0;i<nmsg;i++) {
        s=p;
        p+=sprintf(p,"$GPGSV,%d,%d,%02d",nmsg,i+1,n);
        
        for (j=0;j<4;j++,k++) {
            if (k<n) {
                azs=az[index[k]]*R2D; if (azs<0.0) azs+=360.0;
                p+=sprintf(p,",%02d,%02.0f,%03.0f,%02d",sat[index[k]],
                           el[index[k]]*R2D,azs,snr[index[k]]);
            }
            else p+=sprintf(p,",,,,");
        }
        for (q=s+1,sum=0;*q;q++) sum^=*q; /* check-sum */
        p+=sprintf(p,"*%02X%c%c",sum,0x0D,0x0A);
    }
    return p-(char *)buff;
}
/* output solution header ------------------------------------------------------
* output solution header to buffer
* args   : unsigned char *buff IO output buffer
* return : number of output bytes
* notes  : before calling that set solution options with setsolopt()
*-----------------------------------------------------------------------------*/
extern int outsolheads(unsigned char *buff)
{
    char *s1[]={"WGS84","Tokyo"},*s2[]={"ellipsoidal","geodetic"};
    char *s3[]={"GPST","UTC ","JST "},*sep=opt.sep;
    char *p=(char *)buff;
   
    trace(3,"outsolhead:\n");
    
    if (opt.posf==SOLF_NMEA) return 0;
    
    if (opt.outhead) {
        p+=sprintf(p,"%s (",COMMENTH);
        if      (opt.posf==SOLF_XYZ) p+=sprintf(p,"x/y/z-ecef=WGS84");
        else if (opt.posf==SOLF_ENU) p+=sprintf(p,"e/n/u-baseline=WGS84");
        else p+=sprintf(p,"lat/lon/height=%s/%s",s1[opt.datum],s2[opt.height]);
        p+=sprintf(p,",Q=1:fix,2:float,4:dgps,5:single,ns=# of satellites)\n");
    }
    p+=sprintf(p,"%s  %-*s%s",COMMENTH,(opt.timef?16:8)+(opt.timeu<=0?0:opt.timeu+1),
               s3[opt.times],sep);
    
    if (opt.posf==SOLF_LLH) { /* lat/lon/hgt */
        if (opt.degf) {
            p+=sprintf(p,"%16s%s%16s%s%10s%s%3s%s%3s%s%8s%s%8s%s%8s%s%8s%s%8s%s%8s\n",
                       "latitude(d'\")",sep,"longitude(d'\")",sep,"height(m)",sep,
                       "Q",sep,"ns",sep,"sdn(m)",sep,"sde(m)",sep,"sdu(m)",sep,
                       "sdne(m)",sep,"sdeu(m)",sep,"sdue(m)");
        }
        else {
            p+=sprintf(p,"%14s%s%14s%s%10s%s%3s%s%3s%s%8s%s%8s%s%8s%s%8s%s%8s%s%8s\n",
                       "latitude(deg)",sep,"longitude(deg)",sep,"height(m)",sep,
                       "Q",sep,"ns",sep,"sdn(m)",sep,"sde(m)",sep,"sdu(m)",sep,
                       "sdne(m)",sep,"sdeu(m)",sep,"sdun(m)");
        }
    }
    else if (opt.posf==SOLF_XYZ) { /* x/y/z-ecef */
        p+=sprintf(p,"%14s%s%14s%s%14s%s%3s%s%3s%s%8s%s%8s%s%8s%s%8s%s%8s%s%8s\n",
                   "x-ecef(m)",sep,"y-ecef(m)",sep,"z-ecef(m)",sep,"Q",sep,"ns",sep,
                   "sdx(m)",sep,"sdy(m)",sep,"sdz(m)",sep,"sdxy(m)",sep,
                   "sdyz(m)",sep,"sdzx(m)");
    }
    else if (opt.posf==SOLF_ENU) { /* e/n/u-baseline */
        p+=sprintf(p,"%14s%s%14s%s%14s%s%3s%s%3s%s%8s%s%8s%s%8s%s%8s%s%8s%s%8s\n",
                   "e-baseline(m)",sep,"n-baseline(m)",sep,"u-baseline(m)",sep,
                   "Q",sep,"ns",sep,"sde(m)",sep,"sdn(m)",sep,"sdu(m)",sep,
                   "sden(m)",sep,"sdnu(m)",sep,"sdue(m)");
    }
    return p-(char *)buff;
}
/* output solution body --------------------------------------------------------
* output solution body to buffer
* args   : unsigned char *buff IO output buffer
*          gtime_t time     I   time
*          double *rr       I   solution position/velocity (m) (ecef) (6x1)
*          double *Qr       I   solution position covariance matrix (m) (3x3)
*          double *rb       I   base station position (m) (ecef) (3x1)
*          int    stat      I   solution status (SOLQ_???)
*          int    ns        I   number of valid satellites
* return : number of output bytes
* notes  : before calling that set solution options with setsolopt()
*-----------------------------------------------------------------------------*/
extern int outsols(unsigned char *buff, gtime_t time, const double *rr,
                   const double *Qr, const double *rb, int stat, int ns)
{
    double gpst;
    int week;
    char s[64],*sep=opt.sep;
    unsigned char *p=buff;
    
    trace(3,"outsols:stat=%d ns=%d\n",stat,ns);
    
    if      (opt.times==TIMES_UTC) time=gpst2utc(time);
    else if (opt.times==TIMES_JST) time=timeadd(gpst2utc(time),9*3600.0);
    if (opt.timef) time2str(time,s,opt.timeu);
    else {
        gpst=time2gpst(time,&week);
        if (86400*7-gpst<0.5/pow(10.0,opt.timeu)) gpst=0.0;
        sprintf(s,"%4d%s%*.*f",week,sep,6+(opt.timeu<=0?0:opt.timeu+1),opt.timeu,gpst);
    }
    switch (opt.posf) {
        case SOLF_LLH:  p+=outpos(p,s,rr,Qr,stat,ns); break;
        case SOLF_XYZ:  p+=outecef(p,s,rr,Qr,stat,ns); break;
        case SOLF_ENU:  p+=outenu(p,s,rr,Qr,rb,stat,ns); break;
        case SOLF_NMEA: p+=outnmea_rmc(p,time,rr,stat);
                        p+=outnmea_gga(p,time,rr,stat,ns); break;
    }
    return p-buff;
}
/* output solution extended ----------------------------------------------------
* output solution exteneded infomation
* args   : unsigned char *buff IO output buffer
*          int    stat      I   solution status (SOLQ_???)
*          int    ns        I   number of satellites
*          int    *sat      I   satellite prn numbers
*          double *az       I   azimuth angle (rad)
*          double *el       I   elevation angle (rad)
*          int    *snr      I   snr (dBHz)
*          int    *vsat     I   valid satellite flag
* return : number of output bytes
* notes  : only support nmea
*-----------------------------------------------------------------------------*/
extern int outsolexs(unsigned char *buff, int stat, int ns, int *sat,
                     double *az, double *el, int *snr, int *vsat)
{
    unsigned char *p=buff;
    
    if (opt.posf==SOLF_NMEA) {
        p+=outnmea_gsa(p,stat,ns,sat,az,el,snr,vsat);
        p+=outnmea_gsv(p,stat,ns,sat,az,el,snr,vsat);
    }
    return p-buff;
}
/* output solution header ------------------------------------------------------
* output solution heade to file
* args   : FILE   *fp       I   output file pointer
* return : none
* notes  : before calling that set solution options with setsolopt()
*-----------------------------------------------------------------------------*/
extern void outsolhead(FILE *fp)
{
    unsigned char buff[MAXSOLH];
    int n;
    
    trace(3,"outsolhead:\n");
    
    n=outsolheads(buff);
    fwrite(buff,n,1,fp);
}
/* output solution body --------------------------------------------------------
* output solution body to file
* args   : FILE   *fp       I   output file pointer
*          gtime_t time     I   time
*          double *rr       I   solution position (m) (ecef)
*          double *Qr       I   solution covariance matrix (m)
*          double *rb       I   base station position (m) (ecef)
*          int    stat      I   solution status flag
*          int    ns        I   number of satellites
* return : none
* notes  : before calling that set solution options with setsolopt()
*-----------------------------------------------------------------------------*/
extern void outsol(FILE *fp, gtime_t time, const double *rr, const double *Qr,
                   const double *rb, int stat, int ns)
{
    unsigned char buff[MAXSOL];
    int n;
    
    trace(3,"outsol   :\n");
    
    n=outsols(buff,time,rr,Qr,rb,stat,ns);
    fwrite(buff,n,1,fp);
}
/* output solution extended ----------------------------------------------------
* output solution exteneded infomation to file
* args   : FILE   *fp       I   output file pointer
*          int    stat      I   solution status (SOLQ_???)
*          int    ns        I   number of satellites
*          int    *sat      I   satellite prn numbers
*          double *az       I   azimuth angle (rad)
*          double *el       I   elevation angle (rad)
*          int    *snr      I   snr (dBHz)
*          int    *vsat     I   valid satellite flag
* return : output size (bytes)
* notes  : only support nmea
*-----------------------------------------------------------------------------*/
extern void outsolex(FILE *fp, int stat, int ns, int *sat, double *az,
                     double *el, int *snr, int *vsat)
{
    unsigned char buff[MAXSOL];
    int n;
    
    trace(3,"outsolex :\n");
    
    n=outsolexs(buff,stat,ns,sat,az,el,snr,vsat);
    fwrite(buff,n,1,fp);
}
/* set solution output options -------------------------------------------------
* set output options of text position format
* args   : solopt_t *solopt I   solution options
* return : none
*-----------------------------------------------------------------------------*/
extern void setsolopt(const solopt_t *solopt)
{
    trace(3,"setsolopt:\n");
    
    opt=*solopt;
    if (!*opt.sep) strcpy(opt.sep," "); else opt.sep[1]='\0'; 
}
/* set solution output format --------------------------------------------------
* set output format
* args   : int    format     I  solution format (SOLF_???)
* return : none
*-----------------------------------------------------------------------------*/
extern void setsolformat(int format)
{
    trace(3,"setsolformat: format=%d\n",format);
    
    opt.posf=format;
}
