/*------------------------------------------------------------------------------
* stream.c : stream data input/output functions
*
*          Copyright (C) 2008-2009 by T.TAKASU, All rights reserved.
*
* options : -DWIN32    use WIN32 API
*
* version : $Revision: 1.1 $ $Date: 2008/12/30 08:40:45 $
* history : 2009/01/16 1.0 new
*-----------------------------------------------------------------------------*/
#include <ctype.h>
#include "rtklib.h"
#ifdef WIN32
#include <winsock2.h>
#else
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

static const char rcsid[]="$Id$";

/* constants -----------------------------------------------------------------*/

#define TINTACT             200         /* period for stream active (ms) */
#define SERIBUFFSIZE        4096        /* serial buffer size (bytes) */
#define TIMETAGH_LEN        64          /* time tag file header length */

#define NTRIP_AGENT         "RTKLIB/" VER_RTKLIB
#define NTRIP_PORT          2101        /* default ntrip port */
#define NTRIP_MAXRSP        32768       /* max size of ntrip response */
#define NTRIP_MAXSTR        256         /* max length of mountpoint string */
#define NTRIP_RSP_OK_CLI    "ICY 200 OK\r\n" /* ntrip response: client */
#define NTRIP_RSP_OK_SVR    "OK\r\n"    /* ntrip response: server */
#define NTRIP_RSP_SRCTBL    "SOURCETABLE 200 OK\r\n" /* ntrip response: source table */
#define NTRIP_RSP_TBLEND    "ENDSOURCETABLE"
#define NTRIP_RSP_HTTP      "HTTP/"     /* ntrip response: http */
#define NTRIP_RSP_ERROR     "ERROR"     /* ntrip response: error */

/* macros --------------------------------------------------------------------*/

#ifdef WIN32
#define dev_t               HANDLE
#define socket_t            SOCKET
#define threadfunc          DWORD WINAPI
#else
#define dev_t               int
#define socket_t            int
#define closesocket         close
#define threadfunc          void *
#endif

/* type definition -----------------------------------------------------------*/

typedef struct {            /* serial control type */
    dev_t dev;              /* serial device */
    int error;              /* error state */
#ifdef WIN32
    int state,wp,rp;        /* state,write/read pointer */
    int buffsize;           /* write buffer size (bytes) */
    HANDLE thread;          /* thread */
    lock_t lock;            /* lock flag */
    unsigned char *buff;    /* write buffer */
#endif
} serial_t;

typedef struct {            /* file control type */
    FILE *fp;               /* file pointer */
    FILE *fp_tag;           /* file pointer of tag file */
    gtime_t time;           /* start time */
    unsigned int tick;      /* start tick */
    unsigned int fpos;      /* current file position */
    lock_t lock;            /* lock flag */
} file_t;

typedef struct {            /* tcp control type */
    int state;              /* state (0:close,1:wait,2:connect) */
    int type;               /* type (0:server,1:client) */
    char saddr[256];        /* server address */
    int port;               /* server port */
    struct sockaddr_in addr; /* server address resolved */
    int tcon;               /* reconnect time (ms) (-1:never,0:now) */
    unsigned int tact;      /* data active tick */
    unsigned int tdis;      /* disconnect tick */
    socket_t sock;          /* socket descriptor */
} tcp_t;

typedef struct {            /* ntrip control type */
    int state;              /* state (0:close,1:wait,2:connect) */
    int type;               /* type (0:server,1:client) */
    int nb;                 /* response buffer size */
    char mntpnt[256];       /* mountpoint */
    char user[256];         /* user */
    char passwd[256];       /* password */
    char str[NTRIP_MAXSTR]; /* string */
    unsigned char buff[NTRIP_MAXRSP]; /* response buffer */
    tcp_t *tcp;             /* tcp control */
} ntrip_t;

/* global variables ----------------------------------------------------------*/

static int toinact  =10000; /* inactive timeout (ms) */
static int ticonnect=10000; /* interval to re-connect (ms) */
static int tirate   =1000;  /* avraging time of data rate (ms) */
static int buffsize =32768; /* receive/send buffer size (bytes) */

/* read/write serial buffer --------------------------------------------------*/
#ifdef WIN32
static int readseribuff(serial_t *serial, unsigned char *buff, int nmax)
{
    int ns;
    
    tracet(5,"readseribuff: dev=%d\n",serial->dev);
    
    lock(&serial->lock);
    for (ns=0;serial->rp!=serial->wp&&ns<nmax;ns++) {
       buff[ns]=serial->buff[serial->rp];
       if (++serial->rp>=serial->buffsize) serial->rp=0;
    }
    unlock(&serial->lock);
    tracet(5,"readseribuff: ns=%d rp=%d wp=%d\n",ns,serial->rp,serial->wp);
    return ns;
}
static int writeseribuff(serial_t *serial, unsigned char *buff, int n)
{
    int ns,wp;
    
    tracet(5,"writeseribuff: dev=%d n=%d\n",serial->dev,n);
    
    lock(&serial->lock);
    for (ns=0;ns<n;ns++) {
        serial->buff[wp=serial->wp]=buff[ns];
        if (++wp>=serial->buffsize) wp=0;
        if (wp!=serial->rp) serial->wp=wp; else break;
    }
    unlock(&serial->lock);
    tracet(5,"writeseribuff: ns=%d rp=%d wp=%d\n",ns,serial->rp,serial->wp);
    return ns;
}
#endif /* WIN32 */

/* write serial thread -------------------------------------------------------*/
#ifdef WIN32
static threadfunc serialthread(void *arg)
{
    serial_t *serial=(serial_t *)arg;
    unsigned char buff[128];
    unsigned int tick;
    DWORD ns;
    int n;
    
    tracet(3,"serialthread:\n");
    
    serial->state=1;
    
    for (;;) {
        tick=tickget();
        while ((n=readseribuff(serial,buff,sizeof(buff)))>0) {
            if (!WriteFile(serial->dev,buff,n,&ns,NULL)) serial->error=1;
        }
        if (!serial->state) break;
        sleepms(10-(int)(tickget()-tick)); /* cycle=10ms */
    }
    free(serial->buff);
    return 0;
}
#endif /* WIN32 */

/* open serial ---------------------------------------------------------------*/
static serial_t *openserial(const char *path, int mode, char *msg)
{
    const int br[]={
        300,600,1200,2400,4800,9600,19200,38400,57600,115200,230400
    };
    serial_t *serial;
    int i,brate=9600,bsize=8,stopb=1;
    char *p,parity='N',dev[128],port[128],fctr[64]="";
#ifdef WIN32
    DWORD error,rw=0,siz=sizeof(COMMCONFIG);
    COMMCONFIG cc={0};
    COMMTIMEOUTS co={MAXDWORD,0,0,0,0}; /* non-block-read */
    char dcb[64]="";
#else
    const speed_t bs[]={
        B300,B600,B1200,B2400,B4800,B9600,B19200,B38400,B57600,B115200,B230400
    };
    struct termios ios={0};
    int rw=0;
#endif
    tracet(3,"openserial: path=%s mode=%d\n",path,mode);
    
    if (!(serial=(serial_t *)malloc(sizeof(serial_t)))) return NULL;
    
    if ((p=strchr(path,':'))) {
        strncpy(port,path,p-path); port[p-path]='\0';
        sscanf(p,":%d:%d:%c:%d:%s",&brate,&bsize,&parity,&stopb,fctr);
    }
    else strcpy(port,path);
    
    for (i=0;i<12;i++) if (br[i]==brate) break;
    if (i>=12) {
        sprintf(msg,"bitrate error (%d)",brate);
        tracet(1,"openserial: %s path=%s\n",msg,path);
        free(serial);
        return NULL;
    }
    parity=toupper(parity);
    
#ifdef WIN32
    sprintf(dev,"\\\\.\\%s",port);
    if (mode&STR_MODE_R) rw|=GENERIC_READ;
    if (mode&STR_MODE_W) rw|=GENERIC_WRITE;
    
    serial->dev=CreateFile(dev,rw,0,0,OPEN_EXISTING,0,NULL);
    if (serial->dev==INVALID_HANDLE_VALUE) {
        sprintf(msg,"device open error (%d)",GetLastError());
        tracet(1,"openserial: %s path=%s\n",msg,path);
        free(serial);
        return NULL;
    }
    if (!GetDefaultCommConfig(port,&cc,&siz)) {
        sprintf(msg,"getconfig error (%d)",GetLastError());
        tracet(1,"openserial: %s\n",msg);
        CloseHandle(serial->dev);
        free(serial);
        return NULL;
    }
    sprintf(dcb,"baud=%d parity=%c data=%d stop=%d",brate,parity,bsize,stopb);
    if (!BuildCommDCB(dcb,&cc.dcb)) {
        sprintf(msg,"buiddcb error (%d)",GetLastError());
        tracet(1,"openserial: %s\n",msg);
        CloseHandle(serial->dev);
        free(serial);
        return NULL;
    }
    if (!strcmp(fctr,"rts")) {
        cc.dcb.fRtsControl=RTS_CONTROL_HANDSHAKE;
    }
    SetCommConfig(serial->dev,&cc,siz); /* ignore error to support novatel */
    SetCommTimeouts(serial->dev,&co);
    ClearCommError(serial->dev,&error,NULL);
    PurgeComm(serial->dev,PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
    
    /* create write thread */
    initlock(&serial->lock);
    serial->state=serial->wp=serial->rp=serial->error=0;
    serial->buffsize=buffsize;
    if (!(serial->buff=(unsigned char *)malloc(buffsize))) {
        CloseHandle(serial->dev);
        free(serial);
        return NULL;
    }
    if (!(serial->thread=CreateThread(NULL,0,serialthread,serial,0,NULL))) {
        sprintf(msg,"serial thread error (%d)",GetLastError());
        tracet(1,"openserial: %s\n",msg);
        CloseHandle(serial->dev);
        free(serial);
        return NULL;
    }
    return serial;
#else
    sprintf(dev,"/dev/%s",port);
    
    if ((mode&STR_MODE_R)&&(mode&STR_MODE_W)) rw=O_RDWR;
    else if (mode&STR_MODE_R) rw=O_RDONLY;
    else if (mode&STR_MODE_W) rw=O_WRONLY;
    
    if ((serial->dev=open(dev,rw|O_NOCTTY|O_NONBLOCK))<0) {
        sprintf(msg,"device open error (%d)",errno);
        tracet(1,"openserial: %s dev=%s\n",msg,dev);
        free(serial);
        return NULL;
    }
    tcgetattr(serial->dev,&ios);
    cfsetospeed(&ios,bs[i]);
    cfsetispeed(&ios,bs[i]);
    ios.c_cflag|=bsize==7?CS7:CS8;
    ios.c_cflag|=parity=='O'?(PARENB|PARODD):(parity=='E'?PARENB:0);
    ios.c_cflag|=stopb==2?CSTOPB:0;
    ios.c_cflag|=!strcmp(fctr,"rts")?CRTSCTS:0;
    tcsetattr(serial->dev,TCSANOW,&ios);
    tcflush(serial->dev,TCIOFLUSH);
    return serial;
#endif
}
/* close serial --------------------------------------------------------------*/
static void closeserial(serial_t *serial)
{
    tracet(3,"closeserial: dev=%d\n",serial->dev);
    
    if (!serial) return;
#ifdef WIN32
    serial->state=0;
    WaitForSingleObject(serial->thread,10000);
    CloseHandle(serial->dev);
    CloseHandle(serial->thread);
#else
    close(serial->dev);
#endif
    free(serial);
}
/* read serial ---------------------------------------------------------------*/
static int readserial(serial_t *serial, unsigned char *buff, int n, char *msg)
{
#ifdef WIN32
    DWORD nr;
#else
    int nr;
#endif
    tracet(4,"readserial: dev=%d n=%d\n",serial->dev,n);
    if (!serial) return 0;
#ifdef WIN32
    if (!ReadFile(serial->dev,buff,n,&nr,NULL)) return 0;
#else
    if ((nr=read(serial->dev,buff,n))<0) return 0;
#endif
    tracet(5,"readserial: exit dev=%d nr=%d\n",serial->dev,nr);
    return nr;
}
/* write serial --------------------------------------------------------------*/
static int writeserial(serial_t *serial, unsigned char *buff, int n, char *msg)
{
    int ns;
    
    tracet(3,"writeserial: dev=%d n=%d\n",serial->dev,n);
    
    if (!serial) return 0;
#ifdef WIN32
    if ((ns=writeseribuff(serial,buff,n))<n) serial->error=1;
#else
    if ((ns=write(serial->dev,buff,n))<0) return 0;
#endif
    tracet(5,"writeserial: exit dev=%d ns=%d\n",serial->dev,ns);
    return ns;
}
/* get state serial ----------------------------------------------------------*/
static int stateserial(serial_t *serial)
{
    return !serial?0:(serial->error?-1:2);
}
/* open file -----------------------------------------------------------------*/
static file_t *openfile(const char *path, int mode, char *msg)
{
    file_t *file,file0={0};
    int moderw=STR_MODE_R|STR_MODE_W;
    char *rw,tagpath[1024]="",tagh[TIMETAGH_LEN+1]="";
    
    tracet(3,"openfile: path=%s mode=%d\n",path,mode);
    
    if (!(mode&moderw)||(mode&moderw)==moderw) {
        sprintf(msg,"file mode error");
        return NULL;
    }
    if (mode&STR_MODE_R) rw="rb"; else rw="wb";
    
    if (!(file=(file_t *)malloc(sizeof(file_t)))) return NULL;
    
    *file=file0;
    file->time=utc2gpst(timeget());
    file->tick=tickget();
    file->fpos=0;
    initlock(&file->lock);
    
    if (!(file->fp=fopen(path,rw))) {
        sprintf(msg,"file open error: %s",path);
        tracet(1,"openfile: %s\n",msg);
        free(file);
        return NULL;
    }
    if (mode&STR_MODE_T) { /* time-tag file */
        sprintf(tagpath,"%s.tag",path);
        
        if (!(file->fp_tag=fopen(tagpath,rw))) {
            sprintf(msg,"tag open error: %s",tagpath);
            tracet(1,"openfile: %s\n",msg);
            fclose(file->fp);
            free(file);
            return NULL;
        }
        if (mode&STR_MODE_R) {
            if (fread(&tagh,1,TIMETAGH_LEN,file->fp_tag)<TIMETAGH_LEN||
                strncmp(tagh,"TIMETAG RTKLIB",14)) {
                sprintf(msg,"tag file error: %s",tagpath);
                tracet(1,"openfile: %s\n",msg);
                fclose(file->fp); fclose(file->fp_tag);
                free(file);
                return NULL;
            }
            fread(&file->time,1,sizeof(file->time),file->fp_tag);
        }
        else {
            sprintf(tagh,"TIMETAG RTKLIB %s",VER_RTKLIB);
            fwrite(&tagh,1,TIMETAGH_LEN,file->fp_tag);
            fwrite(&file->time,1,sizeof(file->time),file->fp_tag);
        }
    }
    return file;
}
/* close file ----------------------------------------------------------------*/
static void closefile(file_t *file)
{
    tracet(3,"closefile: fp=%d\n",file->fp);
    
    if (!file) return;
    fclose(file->fp);
    if (file->fp_tag) fclose(file->fp_tag);
    free(file);
}
/* get state file ------------------------------------------------------------*/
static int statefile(file_t *file)
{
    return file?2:0;
}
/* read file -----------------------------------------------------------------*/
static int readfile(file_t *file, unsigned char *buff, int nmax, char *msg)
{
    unsigned int nr=0,t,tick,fpos;
    gtime_t time;
    
    tracet(4,"readfile: fp=%d nmax=%d\n",file->fp,nmax);
    
    if (!file) return 0;
    
    if (file->fp_tag) {
        t=tickget()-file->tick;
        
        while (fread(&tick,sizeof(tick),1,file->fp_tag)==1&&
               fread(&fpos,sizeof(fpos),1,file->fp_tag)==1) {
            if ((int)(tick-t)<5) continue;
            
            time=timeadd(file->time,(double)t/1000.0);
            time2str(time,msg,0);
            
            if ((int)(fpos-file->fpos)>=nmax) {
               fseek(file->fp,fpos,SEEK_SET);
               file->fpos=fpos;
               return 0;
            }
            nmax=(int)(fpos-file->fpos);
            fseek(file->fp_tag,-(long)sizeof(tick)*2,SEEK_CUR);
            break;
        }
    }
    if (nmax>0) {
        nr=fread(buff,1,nmax,file->fp);
        file->fpos+=nr;
    }
    tracet(5,"readfile: fp=%d nr=%d fpos=%d\n",file->fp,nr,file->fpos);
    return (int)nr;
}
/* write file ----------------------------------------------------------------*/
static int writefile(file_t *file, unsigned char *buff, int n, char *msg)
{
    unsigned int ns,tick=tickget();
    
    tracet(3,"writefile: fp=%d n=%d\n",file->fp,n);
    
    if (!file) return 0;
    
    ns=fwrite(buff,1,n,file->fp);
    
    if (file->fp_tag) {
        tick-=file->tick;
        fwrite(&tick,1,sizeof(tick),file->fp_tag);
        fwrite(&file->fpos,1,sizeof(file->fpos),file->fp_tag);
    }
    tracet(5,"writefile: fp=%d ns=%d tick=%5d fpos=%d\n",file->fp,ns,tick,file->fpos);
    
    file->fpos+=ns;
    return (int)ns;
}
/* get socket error ----------------------------------------------------------*/
#ifdef WIN32
static int errsock(void) {return WSAGetLastError();}
#else
static int errsock(void) {return errno;}
#endif

/* set socket option ---------------------------------------------------------*/
static int setsock(tcp_t *tcp, char *msg)
{
    int bs=buffsize,mode=1;
#ifdef WIN32
    int tv=0;
#else
    struct timeval tv={0};
#endif
    tracet(3,"setsock: sock=%d\n",tcp->sock);
    
    if (setsockopt(tcp->sock,SOL_SOCKET,SO_RCVTIMEO,(const char *)&tv,sizeof(tv))==-1||
        setsockopt(tcp->sock,SOL_SOCKET,SO_SNDTIMEO,(const char *)&tv,sizeof(tv))==-1) {
        sprintf(msg,"sockopt error: notimeo");
        tracet(1,"setsock: setsockopt error 1 sock=%d err=%d\n",tcp->sock,errsock());
        closesocket(tcp->sock);
        return 0;
    }
    if (setsockopt(tcp->sock,SOL_SOCKET,SO_RCVBUF,(const char *)&bs,sizeof(bs))==-1||
        setsockopt(tcp->sock,SOL_SOCKET,SO_SNDBUF,(const char *)&bs,sizeof(bs))==-1) {
        tracet(1,"setsock: setsockopt error 2 sock=%d err=%d bs=%d\n",tcp->sock,errsock(),bs);
        sprintf(msg,"sockopt error: bufsiz");
    }
    if (setsockopt(tcp->sock,IPPROTO_TCP,TCP_NODELAY,(const char *)&mode,sizeof(mode))==-1) {
        tracet(1,"setsock: setsockopt error 3 sock=%d err=%d\n",tcp->sock,errsock());
        sprintf(msg,"sockopt error: nodelay");
    }
    return 1;
}
/* generate socket -----------------------------------------------------------*/
static int gensock(tcp_t *tcp, char *msg)
{
    struct hostent *hp;
    
    tracet(3,"gensock: type=%d\n",tcp->type);
    
    /* generate socket */
    if ((tcp->sock=socket(AF_INET,SOCK_STREAM,0))==-1) {
        sprintf(msg,"socket error (%d)",errsock());
        tracet(1,"gensock: socket error err=%d\n",errsock());
        tcp->state=-1;
        return 0;
    }
    if (!setsock(tcp,msg)) {
        tcp->state=-1;
        return 0;
    }
    memset(&tcp->addr,0,sizeof(tcp->addr));
    tcp->addr.sin_family=AF_INET;
    tcp->addr.sin_port=htons(tcp->port);
    
    if (tcp->type==0) { /* server socket */
        if (bind(tcp->sock,(struct sockaddr *)&tcp->addr,sizeof(tcp->addr))==-1) {
            sprintf(msg,"bind error (%d) : %d",errsock(),tcp->port);
            tracet(1,"gensock: bind error port=%d err=%d\n",tcp->port,errsock());
            closesocket(tcp->sock);
            tcp->state=-1;
            return 0;
        }
        listen(tcp->sock,5);
    }
    else { /* client socket */
        if (!(hp=gethostbyname(tcp->saddr))) {
            sprintf(msg,"address error (%s)",tcp->saddr);
            tracet(1,"gensock: gethostbyname error addr=%s err=%d\n",tcp->saddr,errsock());
            tcp->state=1;
            closesocket(tcp->sock);
            tcp->state=-1;
            return 0;
        }
        memcpy(&tcp->addr.sin_addr,hp->h_addr,hp->h_length);
    }
    tcp->state=1;
    tcp->tact=tickget();
    tracet(5,"gensock: exit sock=%d\n",tcp->sock);
    return 1;
}
/* accept client connection ---------------------------------------------------*/
static int accsock(tcp_t *tcp, char *msg)
{
    struct timeval tv={0};
    struct sockaddr_in addr;
    socket_t sock;
    fd_set rs;
    int err,len=sizeof(addr);
    
    tracet(3,"accsock: sock=%d\n",tcp->sock);
    
    /* non-block accept */
    FD_ZERO(&rs); FD_SET(tcp->sock,&rs);
    if (!select(tcp->sock+1,&rs,NULL,NULL,&tv)) {
        sprintf(msg,"waiting...");
        return 0;
    }
    if ((sock=accept(tcp->sock,(struct sockaddr *)&addr,&len))==-1) {
        err=errsock();
        sprintf(msg,"accept error (%d)",err);
        tracet(1,"accsock: accept error sock=%d err=%d\n",tcp->sock,err);
        closesocket(tcp->sock); tcp->state=0;
        return 0;
    }
    closesocket(tcp->sock);
    tcp->sock=sock;
    if (!setsock(tcp,msg)) return 0;
    memcpy(&tcp->addr,&addr,sizeof(addr));
    strcpy(tcp->saddr,inet_ntoa(addr.sin_addr));
    sprintf(msg,"connect tcp (%s)",tcp->saddr);
    tracet(2,"accsock: connected sock=%d addr=%s\n",tcp->sock,tcp->saddr);
    tcp->state=2;
    tcp->tact=tickget();
    return 1;
}
/* connect server ------------------------------------------------------------*/
static int consock(tcp_t *tcp, char *msg)
{
    int err;
#ifdef WIN32
    u_long mode=1; 
#else
    struct timeval tv={0};
    fd_set rs,ws;
    int flag;
#endif
    tracet(3,"consock: sock=%d\n",tcp->sock);
    
    /* wait re-connect */
    if (tcp->tcon<0||(tcp->tcon>0&&(int)(tickget()-tcp->tdis)<tcp->tcon)) {
        return 0;
    }
    /* non-block connect */
#ifdef WIN32
    ioctlsocket(tcp->sock,FIONBIO,&mode);
    if (connect(tcp->sock,(struct sockaddr *)&tcp->addr,sizeof(tcp->addr))==-1) {
        err=errsock();
        
        if (err==WSAEWOULDBLOCK||err==WSAEINPROGRESS||err==WSAEALREADY) {
            sprintf(msg,"connecting...");
            return 0;
        }
        if (err!=WSAEISCONN) {
            sprintf(msg,"connect error (%d)",err);
            tracet(1,"consock: connect error sock=%d err=%d\n",tcp->sock,err);
            closesocket(tcp->sock); tcp->state=0;
            return 0;
        }
    }
#else
    flag=fcntl(tcp->sock,F_GETFL,0);
    fcntl(tcp->sock,F_SETFL,flag|O_NONBLOCK);
    if (connect(tcp->sock,(struct sockaddr *)&tcp->addr,sizeof(tcp->addr))==-1) {
        err=errsock();
        if (err!=EINPROGRESS&&err!=EALREADY) {
            sprintf(msg,"connect error (%d)",err);
            tracet(1,"consock: connect error sock=%d err=%d\n",tcp->sock,err);
            closesocket(tcp->sock); tcp->state=0;
            return 0;
        }
        FD_ZERO(&rs); FD_SET(tcp->sock,&rs); ws=rs;
        if (select(tcp->sock+1,&rs,&ws,NULL,&tv)==0) {
            sprintf(msg,"connecting...");
            return 0;
        }
    }
#endif
    sprintf(msg,"connect tcp (%s)",tcp->saddr);
    tracet(2,"consock: connected sock=%d addr=%s\n",tcp->sock,tcp->saddr);
    tcp->state=2;
    tcp->tact=tickget();
    return 1;
}
/* disconnect tcp ------------------------------------------------------------*/
static void discontcp(tcp_t *tcp, int tcon)
{
    tracet(3,"discontcp: sock=%d tcon=%d\n",tcp->sock,tcon);
    
    closesocket(tcp->sock);
    tcp->state=0;
    tcp->tcon=tcon;
    tcp->tdis=tickget();
}
/* decode path (path=[user[:passwd]@]addr[:port][/mntpnt[:str]]) -------------*/
static void decodepath(const char *path, char *addr, char *port, char *user,
                       char *passwd, char *mntpnt, char *str)
{
    char buff[MAXSTRPATH],*p,*q;
    
    tracet(4,"decodepath: path=%s\n",path);
    
    strcpy(buff,path);
    
    if ((p=strchr(buff,'/'))) {
        if ((q=strchr(p+1,':'))) {
            *q='\0'; if (str) strcpy(str,q+1);
        }
        *p='\0'; if (mntpnt) strcpy(mntpnt,p+1);
    }
    if ((p=strchr(buff,'@'))) {
        *p++='\0';
        if ((q=strchr(buff,':'))) {
             *q='\0'; if (passwd) strcpy(passwd,q+1);
        }
        *q='\0'; if (user) strcpy(user,buff); 
    }
    else p=buff;
    if ((q=strchr(p,':'))) {
        *q='\0'; if (port) strcpy(port,q+1); 
    }
    if (addr) strcpy(addr,p);
}
/* open tcp ------------------------------------------------------------------*/
static tcp_t *opentcp(const char *path, int type, char *msg)
{
    tcp_t *tcp,tcp0={0};
    char port[256]="";
    
    tracet(3,"opentcp: path=%s type=%d\n",path,type);
    
    if (!(tcp=(tcp_t *)malloc(sizeof(tcp_t)))) return NULL;
    *tcp=tcp0;
    decodepath(path,tcp->saddr,port,NULL,NULL,NULL,NULL);
    if (sscanf(port,"%d",&tcp->port)<1) {
        sprintf(msg,"port error: %s",port);
        tracet(1,"opentcp: port error port=%s\n",port);
        free(tcp);
        return NULL;
    }
    tcp->type=type;
    tcp->tcon=0;
    return tcp;
}
/* close tcp -----------------------------------------------------------------*/
static void closetcp(tcp_t *tcp)
{
    tracet(3,"closetcp: sock=%d\n",tcp->sock);
    
    closesocket(tcp->sock);
    free(tcp);
}
/* wait socket accept/connect ------------------------------------------------*/
static int waittcp(tcp_t *tcp, char *msg)
{
    tracet(4,"waittcp: sock=%d state=%d\n",tcp->sock,tcp->state);
    
    if (tcp->state<0) return 0;
    
    if (tcp->state==0) { /* close */
        if (!gensock(tcp,msg)) return 0;
    }
    if (tcp->state==1) { /* wait */
        if (tcp->type==0?!accsock(tcp,msg):!consock(tcp,msg)) return 0;
    }
    if (tcp->state==2) { /* connect */
        if (toinact>0&&(int)(tickget()-tcp->tact)>toinact) {
            sprintf(msg,"inactive timeout");
            tracet(2,"waittcp: inactive timeout sock=%d\n",tcp->sock);
            discontcp(tcp,ticonnect);
            return 0;
        }
    }
    return 1;
}
/* read tcp ------------------------------------------------------------------*/
static int readtcp(tcp_t *tcp, unsigned char *buff, int n, char *msg)
{
    int nr,err;
    
    tracet(4,"readtcp: sock=%d state=%d n=%d\n",tcp->sock,tcp->state,n);
    
    if (!waittcp(tcp,msg)) return 0;
    
    if ((nr=recv(tcp->sock,buff,n,0))==-1) {
        err=errsock();
#ifdef WIN32
        if (err!=WSAEWOULDBLOCK) {
#else
        if (err!=EWOULDBLOCK) {
#endif
            tracet(1,"readtcp: recv error sock=%d err=%d\n",tcp->sock,err);
            sprintf(msg,"disconnect_r (%d)",err);
            discontcp(tcp,ticonnect);
        }
        return 0;
    }
    if (nr>0) tcp->tact=tickget();
    tracet(5,"readtcp: exit sock=%d nr=%d\n",tcp->sock,nr);
    return nr;
}
/* write tcp -----------------------------------------------------------------*/
static int writetcp(tcp_t *tcp, unsigned char *buff, int n, char *msg)
{
    int ns,err;
    
    tracet(3,"writetcp: sock=%d state=%d n=%d\n",tcp->sock,tcp->state,n);
    
    if (!waittcp(tcp,msg)) return 0;
    
    if ((ns=send(tcp->sock,buff,n,0))==-1) {
        err=errsock();
#ifdef WIN32
        if (err!=WSAEWOULDBLOCK) {
#else
        if (err!=EWOULDBLOCK) {
#endif
            tracet(1,"writetcp: send error sock=%d err=%d\n",tcp->sock,err);
            sprintf(msg,"disconnect_w (%d)",err);
            discontcp(tcp,ticonnect);
        }
        return 0;
    }
    if (ns>0) tcp->tact=tickget();
    tracet(5,"writetcp: exit sock=%d ns=%d\n",tcp->sock,ns);
    return ns;
}
/* get state tcp -------------------------------------------------------------*/
static int statetcp(tcp_t *tcp)
{
    return tcp?tcp->state:0;
}
/* base64 encoder ------------------------------------------------------------*/
static int encbase64(char *str, const unsigned char *byte, int n)
{
    const char table[]=
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    int i,j,k,b;
    
    tracet(4,"encbase64: n=%d\n",n);
    
    for (i=j=0;i/8<n;) {
        for (k=b=0;k<6;k++,i++) {
            b<<=1; if (i/8<n) b|=(byte[i/8]>>(7-i%8))&0x1;
        }
        str[j++]=table[b];
    }
    while (j&0x3) str[j++]='=';
    str[j]='\0';
    tracet(5,"encbase64: str=%s\n",str);
    return j;
}
/* send ntrip server request -------------------------------------------------*/
static int reqntrip_s(ntrip_t *ntrip, char *msg)
{
    char buff[256+NTRIP_MAXSTR],*p=buff;
    
    tracet(3,"reqntrip_s: state=%d\n",ntrip->state);
    
    p+=sprintf(p,"SOURCE %s %s\r\n",ntrip->passwd,ntrip->mntpnt);
    p+=sprintf(p,"Source-Agent: NTRIP %s\r\n",NTRIP_AGENT);
    p+=sprintf(p,"STR: %s\r\n",ntrip->str);
    p+=sprintf(p,"\r\n");
    
    if (writetcp(ntrip->tcp,(unsigned char *)buff,p-buff,msg)!=p-buff) return 0;
    tracet(2,"reqntrip_s: send request state=%d ns=%d\n",ntrip->state,p-buff);
    ntrip->state=1;
    return 1;
}
/* send ntrip client request -------------------------------------------------*/
static int reqntrip_c(ntrip_t *ntrip, char *msg)
{
    char buff[1024],user[512],*p=buff;
    
    tracet(3,"reqntrip_c: state=%d\n",ntrip->state);
    
    p+=sprintf(p,"GET /%s HTTP/1.0\r\n",ntrip->mntpnt);
    p+=sprintf(p,"User-Agent: NTRIP %s\r\n",NTRIP_AGENT);
    if (!*ntrip->user) {
        p+=sprintf(p,"Accept: */*\r\n");
        p+=sprintf(p,"Connection: close\r\n");
    }
    else {
        sprintf(user,"%s:%s",ntrip->user,ntrip->passwd);
        p+=sprintf(p,"Authorization: Basic ");
        p+=encbase64(p,(unsigned char *)user,strlen(user));
        p+=sprintf(p,"\r\n");
    }
    p+=sprintf(p,"\r\n");
    
    if (writetcp(ntrip->tcp,(unsigned char *)buff,p-buff,msg)!=p-buff) return 0;
    tracet(2,"reqntrip_c: send request state=%d ns=%d\n",ntrip->state,p-buff);
    ntrip->state=1;
    return 1;
}
/* test ntrip server response ------------------------------------------------*/
static int rspntrip_s(ntrip_t *ntrip, char *msg)
{
    int i;
    char *p,*q;
    
    tracet(3,"rspntrip_s: state=%d nb=%d\n",ntrip->state,ntrip->nb);
    
    if ((p=strstr((char *)ntrip->buff,NTRIP_RSP_OK_SVR))) { /* ok */
        q=(char *)ntrip->buff;
        p+=strlen(NTRIP_RSP_OK_SVR);
        ntrip->nb-=p-q;
        for (i=0;i<ntrip->nb;i++) *q++=*p++;
        ntrip->state=2;
        sprintf(msg,"connect ntrip (%s)",ntrip->tcp->saddr);
        tracet(2,"rspntrip_s: response ok nb=%d\n",ntrip->nb);
        return 1;
    }
    else if ((p=strstr((char *)ntrip->buff,NTRIP_RSP_ERROR))) { /* error */
        strncpy(msg,(char *)ntrip->buff,ntrip->nb); msg[ntrip->nb]=0;
        tracet(1,"rspntrip_s: %s nb=%d\n",msg,ntrip->nb);
        ntrip->state=-1;
        discontcp(ntrip->tcp,-1);
    }
    else if (ntrip->nb>=NTRIP_MAXRSP) { /* buffer overflow */
        sprintf(msg,"response overflow");
        tracet(1,"rspntrip_s: response overflow nb=%d\n",ntrip->nb);
        ntrip->state=-1;
        discontcp(ntrip->tcp,-1);
    }
    tracet(5,"rspntrip_s: exit state=%d nb=%d\n",ntrip->state,ntrip->nb);
    return 0;
}
/* test ntrip client response ------------------------------------------------*/
static int rspntrip_c(ntrip_t *ntrip, char *msg)
{
    int i;
    char *p,*q;
    
    tracet(3,"rspntrip_c: state=%d nb=%d\n",ntrip->state,ntrip->nb);
    
    if ((p=strstr((char *)ntrip->buff,NTRIP_RSP_OK_CLI))) { /* ok */
        q=(char *)ntrip->buff;
        p+=strlen(NTRIP_RSP_OK_CLI);
        ntrip->nb-=p-q;
        for (i=0;i<ntrip->nb;i++) *q++=*p++;
        ntrip->state=2;
        sprintf(msg,"connect ntrip (%s)",ntrip->tcp->saddr);
        tracet(2,"rspntrip_c: response ok nb=%d\n",ntrip->nb);
        return 1;
    }
    if ((p=strstr((char *)ntrip->buff,NTRIP_RSP_SRCTBL))) { /* source table */
        if (!*ntrip->mntpnt) { /* source table request */
            ntrip->state=2;
            sprintf(msg,"source table received");
            tracet(2,"rspntrip_c: receive source table nb=%d\n",ntrip->nb);
            return 1;
        }
        sprintf(msg,"no mountp. reconnect...");
        tracet(2,"rspntrip_c: no mount point nb=%d\n",ntrip->nb);
        ntrip->state=0;
        discontcp(ntrip->tcp,ticonnect); /* re-connect after interval */
    }
    else if ((p=strstr((char *)ntrip->buff,NTRIP_RSP_HTTP))) { /* http response */
        if ((q=strchr(p,'\r'))) *q='\0'; else ntrip->buff[128]='\0';
        strcpy(msg,p);
        tracet(1,"rspntrip_s: %s nb=%d\n",msg,ntrip->nb);
        ntrip->state=-1;
        discontcp(ntrip->tcp,-1);
    }
    else if (ntrip->nb>=NTRIP_MAXRSP) { /* buffer overflow */
        sprintf(msg,"response overflow");
        tracet(1,"rspntrip_s: response overflow nb=%d\n",ntrip->nb);
        ntrip->state=-1;
        discontcp(ntrip->tcp,-1);
    }
    tracet(5,"rspntrip_c: exit state=%d nb=%d\n",ntrip->state,ntrip->nb);
    return 0;
}
/* wait ntrip request/response -----------------------------------------------*/
static int waitntrip(ntrip_t *ntrip, char *msg)
{
    int n;
    char *p;
    
    tracet(4,"waitntrip: state=%d nb=%d\n",ntrip->state,ntrip->nb);
    
    if (ntrip->state<0) return 0; /* error */
    
    if (ntrip->tcp->state<2) ntrip->state=0; /* tcp disconnected */
    
    if (ntrip->state==0) { /* send request */
        if (!(ntrip->type==0?reqntrip_s(ntrip,msg):reqntrip_c(ntrip,msg))) {
            return 0;
        }
        tracet(2,"waitntrip: state=%d nb=%d\n",ntrip->state,ntrip->nb);
    }
    if (ntrip->state==1) { /* read response */
        p=(char *)ntrip->buff+ntrip->nb;
        if ((n=readtcp(ntrip->tcp,(unsigned char *)p,NTRIP_MAXRSP-ntrip->nb-1,msg))==0) {
            tracet(5,"waitntrip: readtcp n=%d\n",n);
            return 0;
        }
        ntrip->nb+=n; ntrip->buff[ntrip->nb]='\0';
        
        /* wait response */
        return ntrip->type==0?rspntrip_s(ntrip,msg):rspntrip_c(ntrip,msg);
    }
    return 1;
}
/* open ntrip ----------------------------------------------------------------*/
static ntrip_t *openntrip(const char *path, int type, char *msg)
{
    ntrip_t *ntrip,ntrip0={0};
    char addr[256]="",port[256]="",tpath[1024];
    
    tracet(3,"openntrip: path=%s type=%d\n",path,type);
    
    if (!(ntrip=(ntrip_t *)malloc(sizeof(ntrip_t)))) return NULL;
    
    *ntrip=ntrip0;
    ntrip->type=type;
    
    /* decode path */
    decodepath(path,addr,port,ntrip->user,ntrip->passwd,ntrip->mntpnt,ntrip->str);
    if (*port=='\0') sprintf(port,"%d",NTRIP_PORT);
    sprintf(tpath,"%s:%s",addr,port);
    
    /* open tcp client stream */
    if (!(ntrip->tcp=opentcp(tpath,1,msg))) {
        tracet(1,"openntrip: opentcp error\n");
        free(ntrip);
        return NULL;
    }
    return ntrip;
}
/* close ntrip ---------------------------------------------------------------*/
static void closentrip(ntrip_t *ntrip)
{
    tracet(3,"closentrip: state=%d\n",ntrip->state);
    
    closetcp(ntrip->tcp);
    free(ntrip);
}
/* read ntrip ----------------------------------------------------------------*/
static int readntrip(ntrip_t *ntrip, unsigned char *buff, int n, char *msg)
{
    int nb;
    
    tracet(4,"readntrip: n=%d\n",n);
    
    if (!waitntrip(ntrip,msg)) return 0;
    if (ntrip->nb>0) { /* read response buffer first */
        nb=ntrip->nb<=n?ntrip->nb:n;
        memcpy(buff,ntrip->buff+ntrip->nb-nb,nb);
        ntrip->nb=0;
        return nb;
    }
    return readtcp(ntrip->tcp,buff,n,msg);
}
/* write ntrip ---------------------------------------------------------------*/
static int writentrip(ntrip_t *ntrip, unsigned char *buff, int n, char *msg)
{
    tracet(3,"writentrip: n=%d\n",n);
    
    if (!waitntrip(ntrip,msg)) return 0;
    return writetcp(ntrip->tcp,buff,n,msg);
}
/* get state ntrip -----------------------------------------------------------*/
static int statentrip(ntrip_t *ntrip)
{
    return !ntrip?0:(ntrip->state==0?ntrip->tcp->state:ntrip->state);
}
/* initialize stream environment -----------------------------------------------
* initialize stream environment
* args   : none
* return : none
*-----------------------------------------------------------------------------*/
extern void strinitcom(void)
{
#ifdef WIN32
    WSADATA data;
#endif
    tracet(3,"strinitcom:\n");

#ifdef WIN32
    WSAStartup(MAKEWORD(2,0),&data);
#endif
}
/* initialize stream -----------------------------------------------------------
* initialize stream struct
* args   : stream_t *stream IO  stream
* return : none
*-----------------------------------------------------------------------------*/
extern void strinit(stream_t *stream)
{
    stream_t stream0={0};
    
    tracet(3,"strinit:\n");
    
    *stream=stream0;
    initlock(&stream->lock);
}
/* open stream -----------------------------------------------------------------
* open stream for read or write
* args   : stream_t *stream IO  stream
*          int type         I   stream type (STR_SERIAL,STR_FILE,STR_TCPSVR,...)
*          int mode         I   stream mode (STR_MODE_???)
*          int path         I   stream path
*              STR_SERIAL  : port[:brate[:bsize[:parity[:stopb[:fctr]]]]]
*                (brate=bps,bsize=7|8,parity=n|o|e,stopb=1|2,fctr=off|rts)
*              STR_FILE    : file path
*              STR_TCPSVR  : :port
*              STR_TCPCLI  : addr:port
*              NTRIPSVR    : user[:passwd]@addr[:port]/mntpnt[:str]
*              NTRIPCLI    : [user[:passwd]]@addr[:port][/mntpnt]
* return : status (0:error,1:ok)
*-----------------------------------------------------------------------------*/
extern int stropen(stream_t *stream, int type, int mode, const char *path)
{
    tracet(3,"stropen: type=%d mode=%d path=%s\n",type,mode,path);
    
    stream->type=type;
    stream->mode=mode;
    strcpy(stream->path,path);
    stream->inb=stream->inr=stream->outb=stream->outr=0;
    stream->tick=tickget();
    stream->inbt=stream->outbt=0;
    stream->msg[0]='\0';
    stream->port=NULL;
    switch (type) {
        case STR_SERIAL  : stream->port=openserial(path,mode,stream->msg); break;
        case STR_FILE    : stream->port=openfile(path,mode,stream->msg);   break;
        case STR_TCPSVR  : stream->port=opentcp(path,0,stream->msg);       break;
        case STR_TCPCLI  : stream->port=opentcp(path,1,stream->msg);       break;
        case STR_NTRIPSVR: stream->port=openntrip(path,0,stream->msg);     break;
        case STR_NTRIPCLI: stream->port=openntrip(path,1,stream->msg);     break;
        default: stream->state=0; return 1;
    }
    stream->state=!stream->port?-1:1;
    return stream->port!=NULL;
}
/* close stream ----------------------------------------------------------------
* close stream
* args   : stream_t *stream IO  stream
* return : none
*-----------------------------------------------------------------------------*/
extern void strclose(stream_t *stream)
{
    tracet(3,"strclose: type=%d mode=%d\n",stream->type,stream->mode);
    
    switch (stream->type) {
        case STR_SERIAL  : closeserial((serial_t *)stream->port); break;
        case STR_FILE    : closefile((file_t *)stream->port);     break;
        case STR_TCPSVR  : closetcp((tcp_t *)stream->port);       break;
        case STR_TCPCLI  : closetcp((tcp_t *)stream->port);       break;
        case STR_NTRIPSVR: closentrip((ntrip_t *)stream->port);   break;
        case STR_NTRIPCLI: closentrip((ntrip_t *)stream->port);   break;
    }
    stream->type=0;
    stream->mode=0;
    stream->state=0;
    stream->inr=stream->outr=0;
    stream->msg[0]='\0';
    stream->port=NULL;
}
/* lock/unlock stream ----------------------------------------------------------
* lock/unlock stream
* args   : stream_t *stream I  stream
* return : none
*-----------------------------------------------------------------------------*/
extern void strlock(stream_t *stream) {lock(&stream->lock);}
extern void strunlock(stream_t *stream) {unlock(&stream->lock);}

/* read stream -----------------------------------------------------------------
* read data from stream (unblocked)
* args   : stream_t *stream I  stream
*          unsinged char *buff O data buffer
*          int    n         I  maximum data length
* return : read data length
* notes  : if no data, return immediately with no data.
*-----------------------------------------------------------------------------*/
extern int strread(stream_t *stream, unsigned char *buff, int n)
{
    unsigned int tick;
    char *msg=stream->msg;
    int nr;
    
    tracet(4,"strread: n=%d\n",n);
    
    if (!(stream->mode&STR_MODE_R)||!stream->port) return 0;
    switch (stream->type) {
        case STR_SERIAL  : nr=readserial((serial_t *)stream->port,buff,n,msg); break;
        case STR_FILE    : nr=readfile((file_t *)stream->port,buff,n,msg);   break;
        case STR_TCPSVR  : nr=readtcp((tcp_t *)stream->port,buff,n,msg);     break;
        case STR_TCPCLI  : nr=readtcp((tcp_t *)stream->port,buff,n,msg);     break;
        case STR_NTRIPCLI: nr=readntrip((ntrip_t *)stream->port,buff,n,msg); break;
        default: return 0;
    }
    stream->inb+=nr;
    tick=tickget(); if (nr>0) stream->tact=tick;
    if ((int)(tick-stream->tick)>=tirate) {
        stream->inr=(stream->inb-stream->inbt)*8000/(tick-stream->tick);
        stream->tick=tick; stream->inbt=stream->inb;
    }
    return nr;
}
/* write stream ----------------------------------------------------------------
* write data to stream (unblocked)
* args   : stream_t *stream I   stream
*          unsinged char *buff I data buffer
*          int    n         I   data length
* return : status (0:error,1:ok)
* notes  : write data to buffer and return immediately.
*-----------------------------------------------------------------------------*/
extern int strwrite(stream_t *stream, unsigned char *buff, int n)
{
    unsigned int tick;
    char *msg=stream->msg;
    int ns;
    
    tracet(3,"strwrite: n=%d\n",n);
    
    if (!(stream->mode&(STR_MODE_W|STR_MODE_A))||!stream->port) return 0;
    switch (stream->type) {
        case STR_SERIAL  : ns=writeserial((serial_t *)stream->port,buff,n,msg); break;
        case STR_FILE    : ns=writefile((file_t *)stream->port,buff,n,msg);   break;
        case STR_TCPSVR  : ns=writetcp((tcp_t *)stream->port,buff,n,msg);     break;
        case STR_TCPCLI  : ns=writetcp((tcp_t *)stream->port,buff,n,msg);     break;
        case STR_NTRIPSVR: ns=writentrip((ntrip_t *)stream->port,buff,n,msg); break;
        default: return 0;
    }
    stream->outb+=ns;
    tick=tickget(); if (ns>0) stream->tact=tick;
    if ((int)(tick-stream->tick)>tirate) {
        stream->outr=(stream->outb-stream->outbt)*8000/(tick-stream->tick);
        stream->tick=tick; stream->outbt=stream->outb;
    }
    return ns;
}
/* get stream status -----------------------------------------------------------
* get stream status
* args   : stream_t *stream I   stream
*          char   *msg      IO  status message (NULL: no output)
* return : status (-1:error,0:close,1:wait,2:connect,3:active)
*-----------------------------------------------------------------------------*/
extern int strstat(stream_t *stream, char *msg)
{
    int state;
    
    tracet(4,"strstat:\n");
    
    if (msg) {strncpy(msg,stream->msg,MAXSTRMSG-1); msg[MAXSTRMSG-1]='\0';}
    if (!stream->port) return stream->state;
    switch (stream->type) {
        case STR_SERIAL  : state=stateserial((serial_t *)stream->port); break;
        case STR_FILE    : state=statefile((file_t *)stream->port);     break;
        case STR_TCPSVR  :
        case STR_TCPCLI  : state=statetcp((tcp_t *)stream->port);       break;
        case STR_NTRIPSVR:
        case STR_NTRIPCLI: state=statentrip((ntrip_t *)stream->port);   break;
        default: return 0;
    }
    if (state==2&&(int)(tickget()-stream->tact)<=TINTACT) state=3;
    return state;
}
/* get stream statistics summary -----------------------------------------------
* get stream statistics summary
* args   : stream_t *stream I   stream
*          int    *inb,*inr IO   bytes/rate of input  (NULL: no output)
*          int    *outb,*outr IO bytes/rate of output (NULL: no output)
* return : none
*-----------------------------------------------------------------------------*/
extern void strsum(stream_t *stream, int *inb, int *inr, int *outb, int *outr)
{
    tracet(4,"strsum:\n");
    
    if (inb) *inb=stream->inb;
    if (inr) *inr=stream->inr;
    if (outb) *outb=stream->outb;
    if (outr) *outr=stream->outr;
}
/* set stream options ----------------------------------------------------------
* set stream options
* args   : int    *opt      I   options
*              opt[0]= inactive timeout (ms)
*              opt[1]= interval to reconnect (ms)
*              opt[2]= averaging time of data rate (ms)
*              opt[3]= receive/send buffer size (bytes);
* return : none
*-----------------------------------------------------------------------------*/
extern void strsetopt(const int *opt)
{
    tracet(3,"strsetopt: opt=%d %d %d %d\n",opt[0],opt[1],opt[2],opt[3]);
    
    toinact  =opt[0];
    ticonnect=opt[1];
    tirate   =opt[2];
    buffsize =opt[3];
}
/* stearm server thread ------------------------------------------------------*/
static threadfunc strsvrthread(void *arg)
{
    strsvr_t *svr=(strsvr_t *)arg;
    unsigned int tick;
    int i,n;
    
    tracet(3,"strsvrthread:\n");
    
    svr->state=1;
    svr->tick=tickget();
    
    while (svr->state) {
        tick=tickget();
        
        n=strread(svr->stream,svr->buff,svr->buffsize);
        for (i=1;i<svr->nstr;i++) {
            strwrite(svr->stream+i,svr->buff,n);
        }
        sleepms(svr->cycle-(int)(tickget()-tick));
    }
    for (i=0;i<svr->nstr;i++) strclose(svr->stream+i);
    free(svr->buff);
    
    return 0;
}
/* initialize stream server ----------------------------------------------------
* initialize stream server
* args   : strsvr_t *svr    IO  stream sever struct
*          int    nout      I   number of output streams
* return : none
*-----------------------------------------------------------------------------*/
extern void strsvrinit(strsvr_t *svr, int nout)
{
    strsvr_t svr0={0};
    int i;
    
    tracet(3,"strsvrinit: nout=%d\n",nout);
    
    *svr=svr0;
    for (i=0;i<nout+1&&i<16;i++) strinit(svr->stream+i);
    svr->nstr=i;
}
/* start stream server ---------------------------------------------------------
* start stream server
* args   : strsvr_t *svr    IO  stream sever struct
*          int   *opts      I   stream options
*              opts[0]= inactive timeout (ms)
*              opts[1]= interval to reconnect (ms)
*              opts[2]= averaging time of data rate (ms)
*              opts[3]= receive/send buffer size (bytes);
*              opts[4]= server cycle (ms)
*          int   *strs      I   stream types (STR_???)
*              strs[0]= input stream
*              strs[1]= output stream
*              ...
*          char  **paths    I   stream paths
*          char  *cmd       I   input stream start command (NULL: no cmd)
* return : status (0:error,1:ok)
*-----------------------------------------------------------------------------*/
extern int strsvrstart(strsvr_t *svr, int *opts, int *strs, char **paths,
                       const char *cmd)
{
    int i,rw;
    char cmdend[]="\r\n";
    
    tracet(3,"strsvrstart:\n");
    
    if (svr->state) return 0;
    
    strinitcom();
    strsetopt(opts);
    svr->cycle=opts[4];
    svr->buffsize=opts[3];
    
    if (!(svr->buff=(unsigned char *)malloc(svr->buffsize))) {
        return 0;
    }
    /* open streams */
    for (i=0;i<svr->nstr;i++) {
        if (i>0&&!strcmp(paths[i],paths[0])) {
            sprintf(svr->stream[i].msg,"file open error: %s",paths[i]);
            for (i--;i>=0;i--) strclose(svr->stream+i);
            return 0;
        }
        rw=i==0?STR_MODE_R:STR_MODE_W;
        if (strs[i]!=STR_FILE) rw|=STR_MODE_W;
        if (strs[i]==STR_FILE) rw|=STR_MODE_T;
        if (stropen(svr->stream+i,strs[i],rw,paths[i])) continue;
        for (i--;i>=0;i--) strclose(svr->stream+i);
        return 0;
    }
    /* write start command to input stream */
    if (cmd) {
        strwrite(svr->stream,(unsigned char *)cmd,strlen(cmd));
        strwrite(svr->stream,(unsigned char *)cmdend,2);
    }
    /* create stream server thread */
#ifdef WIN32
    if (!(svr->thread=CreateThread(NULL,0,strsvrthread,svr,0,NULL))) {
#else
    if (pthread_create(&svr->thread,NULL,strsvrthread,svr)) {
#endif
        for (i=0;i<svr->nstr;i++) strclose(svr->stream+i);
        return 0;
    }
    return 1;
}
/* stop stream server ----------------------------------------------------------
* start stream server
* args   : strsvr_t *svr    IO  stream sever struct
*          char  *cmd       I   input stop command (NULL: no cmd)
* return : none
*-----------------------------------------------------------------------------*/
extern void strsvrstop(strsvr_t *svr, const char *cmd)
{
    char cmdend[]="\r\n";
    
    tracet(3,"strsvrstop:\n");
    
    if (cmd) {
        strwrite(svr->stream,(unsigned char *)cmd,strlen(cmd));
        strwrite(svr->stream,(unsigned char *)cmdend,2);
    }
    svr->state=0;
    
#ifdef WIN32
    WaitForSingleObject(svr->thread,10000);
    CloseHandle(svr->thread);
#else
    pthread_join(svr->thread,NULL);
#endif
}
/* get stream server status ----------------------------------------------------
* start stream server
* args   : strsvr_t *svr    IO  stream sever struct
*          int    *stat     O   stream status
*          int    *byte     O   bytes received/sent
*          int    *bps      O   bitrate received/sent
*          char   *msg      O   messages
* return : none
*-----------------------------------------------------------------------------*/
extern void strsvrstat(strsvr_t *svr, int *stat, int *byte, int *bps, char *msg)
{
    char s[MAXSTRMSG]="",*p=msg;
    int i;
    
    tracet(4,"strsvrstat:\n");
    
    for (i=0;i<svr->nstr;i++) {
        if (i==0) {
            strsum(svr->stream,byte,bps,NULL,NULL);
            stat[i]=strstat(svr->stream,s);
        }
        else {
            strsum(svr->stream+i,NULL,NULL,byte+i,bps+i);
            stat[i]=strstat(svr->stream+i,s);
        }
        if (*s) p+=sprintf(p,"(%d) %s ",i,s);
    }
}
